#include "memory.h"
#include "location.h"
#include "state.h"
#include "globals.h"
#include "process.h"
#include "pbl.c"
#include "pblhash.c"
#include "pbl.h"
#include "execution.h"
#include "macros.h"
#include <time.h>
#include "clone.c"
#include "allocate.c"
#include "lvalue.c"
#include "rvalue.c"
#include "value-structures.c"

void start_execution(void) 
{
	struct proc_cfg * main_proc; 
	s_process * current_process;
	int i;
//	while (1) {
		printf ("**********start_execution***********\n");
		state_hash_table = pblHtCreate ();
		if (state_hash_table == NULL)
		{
			sprintf (stderr, "error in allocating state_hash_table\n");
			exit(0);
		}
		/*identify the main procedure in the list of all procedures*/
		seed = int_global_seed;
		main_process = (s_process *) xmalloc (sizeof (s_process));
		printf ("********main process ******\n");
		current_process = NULL;
		main_proc = procs;
		while (1)
		{
			if (main_proc == NULL ||strcmp (main_proc->proc_name, "main") == 0 )
				break;
			main_proc = main_proc->next;
		}
		if (main_proc == NULL)
		{
			fprintf (stderr, "function <main> is not defined\n");
			exit(0);
		}
		curr_proc = main_proc;
		allocate_curr_state ();	
		main_process->process_state = r_curr_state;
		main_process->curr_proc = curr_proc;
		main_process->curr_cfg_node = curr_cfg_node;
		main_process->last_allocated_memory = last_allocated_memory;
		main_process->last_allocated_heap = last_allocated_heap;
		current_process = main_process;
		curr_cfg_node = curr_proc->cfg_array[1];
		execute();
	//}
}

void execute ()
{
	tree curr_insn;
	struct cfg_node * next_cfg_node = NULL;
	s_state * r_prev_state = NULL;
	s_state * r_next_state = NULL;
	s_state * r_other_state = NULL;
	
	s_mem_loc_list  lhs_mem;
	while (curr_cfg_node != NULL) {
		curr_insn = curr_cfg_node->insn;
		if (curr_insn!= NULL && TREE_CODE (curr_insn) == FUNCTION_DECL) {
		
		} else if (curr_insn!= NULL && TREE_CODE (curr_insn) == MODIFY_EXPR) {
			s_value * rhs_value;

			//printf ("modify statements\n");
			if (TREE_CODE (TREE_OPERAND (curr_insn, 1)) == CALL_EXPR) {
				tree formal_params;
				tree actual_params;
				s_value * param_value;
				s_value * global_value;
				char * proc_name;
				struct global_list *temp_globals_list;


				proc_name = IDENTIFIER_POINTER (DECL_NAME (TREE_OPERAND (TREE_OPERAND (TREE_OPERAND (curr_insn,1), 0), 0))); /*FIXME:*/
				//printf ("** %s ** call expression\n", proc_name);
				if (strcmp (proc_name, "malloc") != 0 && strcmp (proc_name, "VS_toss") != 0 && strcmp (proc_name, "fork") != 0) {
					tree temp_insn;
					r_next_state = (s_state *) xmalloc (sizeof (s_state)); 
					clone_state (r_curr_state, r_next_state);
					r_next_state->stmt = curr_insn;
					r_next_state->node = curr_cfg_node;
					r_curr_state = r_next_state;

					temp_insn = TREE_OPERAND (curr_insn, 1);			
					r_prev_state = r_curr_state;
					curr_proc = procs;
					while (1)
					{
						if (curr_proc == NULL ||strcmp (curr_proc->proc_name, proc_name) == 0 )
							break;
						curr_proc = curr_proc->next;
					}
					if (curr_proc == NULL)
					{
						fprintf (stderr, "function <%s> is not defined\n", proc_name);
						exit(0);
					}
					allocate_curr_state ();
					r_curr_state->return_state = r_prev_state;
					formal_params = DECL_ARGUMENTS (curr_proc->proc);
					actual_params = TREE_OPERAND (temp_insn, 1);
					while (formal_params && actual_params) {
						lhs_mem = find_lvalue (r_curr_state, formal_params);
						param_value = find_rvalue (r_prev_state, TREE_VALUE (actual_params));
						lhs_mem->location->val = param_value;
						if (param_value && param_value->kind == HEAP_KIND)
							lhs_mem->location->points_to_index = param_value->value.heap_val->location->index;
						formal_params = TREE_CHAIN (formal_params);
						actual_params = TREE_CHAIN (actual_params);
					}

					temp_globals_list = globals_list;
					while (temp_globals_list != NULL) {
						if (TREE_CODE (TREE_TYPE (temp_globals_list->global)) == RECORD_TYPE) {
							copy_record (r_prev_state, r_curr_state, temp_globals_list->global,0);
							
						} else {
							lhs_mem = find_lvalue (r_curr_state, temp_globals_list->global);		
							global_value = find_rvalue (r_prev_state, temp_globals_list->global);
							lhs_mem->location->val = global_value;
							if (global_value && global_value->kind == HEAP_KIND)
								lhs_mem->location->points_to_index = global_value->value.heap_val->location->index;
						}
						temp_globals_list = temp_globals_list->next;
					}
					clone_heap (r_prev_state, r_curr_state);
					CHECK_HASH_TABLE();
					next_cfg_node = curr_proc->cfg_array[1];
				} else if (strcmp (proc_name, "VS_toss") != 0 && strcmp (proc_name, "fork") != 0) {
					int malloc_size;
					tree type;
					tree temp_insn;

					r_next_state = (s_state *) xmalloc (sizeof (s_state)); 
					clone_state (r_curr_state, r_next_state);
					r_next_state->stmt = curr_insn;
					r_next_state->node = curr_cfg_node;
					r_curr_state = r_next_state;

					temp_insn = TREE_OPERAND (curr_insn, 1);
					lhs_mem = find_lvalue (r_curr_state, TREE_OPERAND (curr_insn, 0));
					malloc_size = TREE_VALUE (TREE_OPERAND (temp_insn, 1))->int_cst.int_cst.low ;
					type = TREE_TYPE (curr_insn);
					rhs_value = create_value_structure_for_pointer (malloc_size, type);
					if (lhs_mem != NULL) {
						if (TREE_CODE (TREE_OPERAND (curr_insn, 0)) == ARRAY_REF) {
							lhs_mem->location->val = rhs_value;
							lhs_mem->location->points_to_index = rhs_value->value.heap_val->location->index;
						} else if (TREE_CODE (TREE_OPERAND (curr_insn, 0)) == INDIRECT_REF) {
							lhs_mem->location->val = rhs_value;
							lhs_mem->location->points_to_index = rhs_value->value.heap_val->location->index;
						} else {
							lhs_mem->location->val = rhs_value;
							lhs_mem->location->points_to_index = rhs_value->value.heap_val->location->index;
						}
					}
					CHECK_HASH_TABLE();
					next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
				} else if (strcmp (proc_name, "fork") != 0) {
					tree temp_insn;
					int range;

					r_next_state = (s_state *) xmalloc (sizeof (s_state)); 
					clone_state (r_curr_state, r_next_state);
					r_next_state->stmt = curr_insn;
					r_next_state->node = curr_cfg_node;
					r_curr_state = r_next_state;

					temp_insn = TREE_OPERAND (curr_insn, 1);
					lhs_mem = find_lvalue (r_curr_state, TREE_OPERAND (curr_insn, 0));
					rhs_value = find_rvalue (r_curr_state, TREE_VALUE (TREE_OPERAND (temp_insn, 1)));
					if (rhs_value != NULL) {
						range = rhs_value->value.int_val;
					} else {
						range = -1;
					}
					rhs_value = (s_value *) xmalloc (sizeof (s_value));
					//srand (seed);
					//seed = seed + 3;
					seed = (unsigned)time (NULL);
					printf ("%d\n", seed);
					srand(seed);
					rhs_value->value.int_val = ((range+1) * (rand()/(RAND_MAX + 1.0))); 
					printf ("VS_toss(%d) = %d\n",range,  rhs_value->value.int_val);
					if (lhs_mem != NULL) {
						if (TREE_CODE (TREE_OPERAND (curr_insn, 0)) == ARRAY_REF) {
							lhs_mem->location->val = rhs_value;
						} else if (TREE_CODE (TREE_OPERAND (curr_insn, 0)) == INDIRECT_REF) {
							lhs_mem->location->val = rhs_value;
						} else {
							lhs_mem->location->val = rhs_value;
						}
					}
					CHECK_HASH_TABLE();
					next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
					
				} else {
					tree temp_insn;
					struct cfg_node * prev_cfg_node;
					s_mem_loc_list prev_last_allocated_memory;
					s_mem_loc_list prev_last_allocated_heap;

					r_next_state = (s_state *) xmalloc (sizeof (s_state)); 
					clone_state (r_curr_state, r_next_state);
					r_next_state->stmt = curr_insn;
					r_next_state->node = curr_cfg_node;
					r_curr_state = r_next_state;

					temp_insn = TREE_OPERAND (curr_insn, 1);
					lhs_mem = find_lvalue (r_curr_state, TREE_OPERAND (curr_insn, 0));
					rhs_value = (s_value *) xmalloc (sizeof (s_value));
					rhs_value->value.int_val = 0;
					if (lhs_mem != NULL) {
						if (TREE_CODE (TREE_OPERAND (curr_insn, 0)) == ARRAY_REF) {
							lhs_mem->location->val = rhs_value;
						} else if (TREE_CODE (TREE_OPERAND (curr_insn, 0)) == INDIRECT_REF) {
							lhs_mem->location->val = rhs_value;
						} else {
							lhs_mem->location->val = rhs_value;
						}
					}
					CHECK_HASH_TABLE();
					next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
					curr_cfg_node = next_cfg_node;
					if (curr_cfg_node != NULL) {
						curr_insn = curr_cfg_node->insn;
					}
					r_prev_state = r_curr_state;
					prev_cfg_node = curr_cfg_node;
					prev_last_allocated_memory = last_allocated_memory;
					prev_last_allocated_heap = last_allocated_heap;

					next_cfg_node = curr_proc->cfg_array[prev_cfg_node->outgoing_edges->node[1]];
					child_process = create_process();
					copy_shared_vars_among_processes(r_prev_state, r_curr_state);
					child_process->process_state =  r_curr_state;
					child_process->process_state->stmt = next_cfg_node->insn;
					child_process->process_state->node = next_cfg_node;
					child_process->curr_proc = curr_proc;
					child_process->curr_cfg_node = next_cfg_node;
					child_process->last_allocated_memory = last_allocated_memory;
					child_process->last_allocated_heap = last_allocated_heap;

					next_cfg_node = curr_proc->cfg_array[prev_cfg_node->outgoing_edges->next->node[1]];
					curr_cfg_node = next_cfg_node;
					main_process->process_state = r_prev_state; 
					main_process->curr_proc = curr_proc;
					main_process->curr_cfg_node = curr_cfg_node;
					main_process->last_allocated_memory = prev_last_allocated_memory;
					main_process->last_allocated_heap = prev_last_allocated_heap;
					current_process = main_process;
					r_curr_state = r_prev_state;
					last_allocated_memory = prev_last_allocated_memory;
					last_allocated_heap = prev_last_allocated_heap;
				}
			} else {
				tree temp_insn;	

				temp_insn = TREE_OPERAND (curr_insn, 1);
				r_next_state = (s_state *) xmalloc (sizeof (s_state)); 
				clone_state (r_curr_state, r_next_state);
				r_next_state->stmt = curr_insn;
				r_next_state->node = curr_cfg_node;
				r_curr_state = r_next_state;
				lhs_mem = find_lvalue (r_curr_state, TREE_OPERAND (curr_insn, 0));
				if (TREE_CODE (temp_insn) == CONVERT_EXPR) {
					rhs_value = find_rvalue (r_curr_state, TREE_OPERAND (temp_insn, 0));
				} else if (TREE_CODE (temp_insn) == NOP_EXPR) {
					rhs_value = find_rvalue (r_curr_state, TREE_OPERAND (temp_insn, 0));
				} else {
					rhs_value = find_rvalue (r_curr_state, temp_insn);
				}
				if (lhs_mem != NULL) {
					if (TREE_CODE (TREE_OPERAND (curr_insn, 0)) == ARRAY_REF) {
						lhs_mem->location->val = rhs_value;
						if (rhs_value->kind == HEAP_KIND)
							lhs_mem->location->points_to_index = rhs_value->value.heap_val->location->index;
					} else if (TREE_CODE (TREE_OPERAND (curr_insn, 0)) == INDIRECT_REF) {
						lhs_mem->location->val = rhs_value;
						if (rhs_value->kind == HEAP_KIND)
							lhs_mem->location->points_to_index = rhs_value->value.heap_val->location->index;
					} else {
						lhs_mem->location->val = rhs_value;
						if (rhs_value->kind == HEAP_KIND)
							lhs_mem->location->points_to_index = rhs_value->value.heap_val->location->index;
						if (rhs_value->kind == MEM_KIND)
							lhs_mem->location->points_to_index = rhs_value->value.mem_val->location->index;
					}
				}
				CHECK_HASH_TABLE();
				next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
			}
			check_rule();
			if (found_second == 1) {
				printf ("property is satisfied\n");
				exit(0);
			}
		} else if (curr_insn!= NULL && TREE_CODE (curr_insn) == COND_EXPR) {
			s_value * cond_value;

			//printf ("conditional statement\n");
			cond_value = find_rvalue (r_curr_state, COND_EXPR_COND (curr_insn));		
			/* cond_value will always have bool_val or int_val */
			if (cond_value->value.int_val == 1 || cond_value->value.bool_val == 1)
				next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
			else
				next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->next->node[1]];
			check_rule();
			if (found_second == 1) {
				printf ("property is satisfied\n");
				exit(0);
			}
		} else if (curr_insn!= NULL && TREE_CODE (curr_insn) == CALL_EXPR) {
			tree formal_params;
			tree actual_params;
			s_value * param_value;
			s_value * global_value;
			char * proc_name;
			struct global_list *temp_globals_list;

			r_next_state = (s_state *) xmalloc (sizeof (s_state)); 
			clone_state (r_curr_state, r_next_state);
			r_next_state->stmt = curr_insn;
			r_next_state->node = curr_cfg_node;
			r_curr_state = r_next_state;
			r_prev_state = r_curr_state;
			proc_name = IDENTIFIER_POINTER (DECL_NAME (TREE_OPERAND (TREE_OPERAND (curr_insn, 0), 0))); /*FIXME:*/
			//printf ("** %s ** call statement\n", proc_name);
			if (strcmp (proc_name, "wait_process") != 0) { 
				curr_proc = procs;
				while (1)
				{
					if (curr_proc == NULL ||strcmp (curr_proc->proc_name, proc_name) == 0 )
						break;
					curr_proc = curr_proc->next;
				}
				if (curr_proc == NULL)
				{
					fprintf (stderr, "function <%s> is not defined\n", proc_name);
					exit(0);
				}
				allocate_curr_state ();
				r_curr_state->return_state = r_prev_state;
				formal_params = DECL_ARGUMENTS (curr_proc->proc);
				actual_params = TREE_OPERAND (curr_insn, 1);
				while (formal_params && actual_params) {
					lhs_mem = find_lvalue (r_curr_state, formal_params);
					param_value = find_rvalue (r_prev_state, TREE_VALUE (actual_params));
					lhs_mem->location->val = param_value;
					if (param_value && param_value->kind == HEAP_KIND)
						lhs_mem->location->points_to_index = param_value->value.heap_val->location->index;
					formal_params = TREE_CHAIN (formal_params);
					actual_params = TREE_CHAIN (actual_params);
				}
				
				temp_globals_list = globals_list;
				while (temp_globals_list != NULL) {
					if (TREE_CODE (TREE_TYPE (temp_globals_list->global)) == RECORD_TYPE) {
						copy_record (r_prev_state, r_curr_state, temp_globals_list->global, 0);
					} else {
						lhs_mem = find_lvalue (r_curr_state, temp_globals_list->global);		
						global_value = find_rvalue (r_prev_state, temp_globals_list->global);
						lhs_mem->location->val = global_value;
						if (global_value && global_value->kind == HEAP_KIND)
							lhs_mem->location->points_to_index = global_value->value.heap_val->location->index;
					}
					temp_globals_list = temp_globals_list->next;
				}
				clone_heap (r_prev_state, r_curr_state);
				CHECK_HASH_TABLE();
				next_cfg_node = curr_proc->cfg_array[1];
				check_rule();
				if (found_second == 1) {
					printf ("property is satisfied\n");
					exit(0);
				}
			} else {
				s_state * prev_state;
				//switch_process ();
				if (current_process == main_process) {
					printf ("********child process ******\n");
					prev_state = r_curr_state;
					main_process->process_state = r_curr_state;
					main_process->curr_proc = curr_proc;
					next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
					main_process->curr_cfg_node = next_cfg_node;
					main_process->last_allocated_memory = last_allocated_memory;
					main_process->last_allocated_heap = last_allocated_heap;
					current_process = child_process;
					curr_cfg_node = current_process->curr_cfg_node;
					curr_proc = current_process->curr_proc;
					r_curr_state = current_process->process_state;
					copy_shared_vars_among_processes(prev_state, r_curr_state);
					last_allocated_memory = current_process->last_allocated_memory;
					last_allocated_heap = current_process->last_allocated_heap;
				}
				else { 
					printf ("********main process ******\n");
					prev_state = r_curr_state;
					child_process->process_state = r_curr_state;
					child_process->curr_proc = curr_proc;
					next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
					child_process->curr_cfg_node = next_cfg_node;
					child_process->last_allocated_memory = last_allocated_memory;
					child_process->last_allocated_heap = last_allocated_heap;
					current_process = main_process;
					curr_cfg_node = current_process->curr_cfg_node;
					curr_proc = current_process->curr_proc;
					r_curr_state = current_process->process_state;
					copy_shared_vars_among_processes(prev_state, r_curr_state);
					last_allocated_memory = current_process->last_allocated_memory;
					last_allocated_heap = current_process->last_allocated_heap;
				}
				next_cfg_node = curr_cfg_node;
				//next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
			}
		} else if (curr_insn!= NULL && TREE_CODE (curr_insn) == RETURN_EXPR) {
			s_value * return_value;
			s_mem_loc_list temp_memory;
			s_mem_loc_list temp_memorya;

			//printf ("return statement\n");
			if (TREE_OPERAND (curr_insn, 0)) {
				curr_insn = TREE_OPERAND (curr_insn, 0);
			}	
			if (TREE_CODE (curr_insn) == MODIFY_EXPR) {
				return_value = find_rvalue (r_curr_state, TREE_OPERAND (curr_insn, 1));	
			} else {
				return_value = find_rvalue (r_curr_state, curr_insn);	
			}
			
			//printf ("proc_name %s\n", curr_proc->proc_name);
			if (strcmp (curr_proc->proc_name, "VS_assert") == 0) {
				//printf ("return value %d\n", return_value->value.int_val);
				if (return_value->value.int_val == 0) {
					printf ("Assertion is not valid at this point\n");
					exit(0);
				}
			}	
			r_prev_state = r_curr_state;
			r_curr_state = r_prev_state->return_state;
			
			/*if return state is null means this is the first procedure in the call graph */
			if (r_curr_state != NULL) {
				struct global_list *temp_globals_list;
				s_value * global_value;
				s_mem_loc_list lhs_mem;

				curr_proc = r_curr_state->proc;
				curr_insn = r_curr_state->stmt;
				curr_cfg_node = r_curr_state->node;
				temp_globals_list = globals_list;
				while (temp_globals_list != NULL) {
					if (TREE_CODE (TREE_TYPE (temp_globals_list->global)) == RECORD_TYPE) {
						copy_record (r_prev_state, r_curr_state, temp_globals_list->global, 0);
					} else {
						lhs_mem = find_lvalue (r_curr_state, temp_globals_list->global);		
						global_value = find_rvalue (r_prev_state, temp_globals_list->global);
						lhs_mem->location->val = global_value;
						if (global_value && global_value->kind == HEAP_KIND)
							lhs_mem->location->points_to_index = global_value->value.heap_val->location->index;
					}
					temp_globals_list = temp_globals_list->next;
				}

				/*
				temp_memory = r_curr_state->memory;	
				while (temp_memory != NULL) {
					temp_memorya = r_prev_state->memory;
					while (temp_memorya != NULL) {
						if (temp_memory->location->var == temp_memorya->location->var) {
							temp_memory->location->index = temp_memorya->location->index;
							temp_memory->location->points_to_index = temp_memorya->location->points_to_index;
							if (temp_memorya->location->val) {
								temp_memory->location->val->value = temp_memorya->location->val->value;
								temp_memory->location->val->kind = temp_memorya->location->val->kind;
							} else {
								temp_memory->location->val = NULL;
							}
							break;
						} else {
							temp_memorya = temp_memorya->next;
						}	
					}
					temp_memory = temp_memory->next;
				}
				*/
				clone_heap (r_prev_state, r_curr_state);
				/* if the target of return is assignment then update the value of lhs by return value*/
				if (TREE_CODE (curr_insn) == MODIFY_EXPR) {
					lhs_mem = find_lvalue (r_curr_state, TREE_OPERAND (curr_insn, 0));
					if (lhs_mem != NULL) {
						if (TREE_CODE (TREE_OPERAND (curr_insn, 0)) == ARRAY_REF) {
							lhs_mem->location->val = return_value;
							if (return_value->kind == HEAP_KIND)
								lhs_mem->location->points_to_index = return_value->value.heap_val->location->index;
						} else if (TREE_CODE (TREE_OPERAND (curr_insn, 0)) == INDIRECT_REF) {
							lhs_mem->location->val = return_value;
							if (return_value->kind == HEAP_KIND)
								lhs_mem->location->points_to_index = return_value->value.heap_val->location->index;
						} else {
							lhs_mem->location->val = return_value;
							if (return_value->kind == HEAP_KIND)
								lhs_mem->location->points_to_index = return_value->value.heap_val->location->index;
						}
					}
				}
				CHECK_HASH_TABLE();
				next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
			} else {
				//printf ("************** values *******************\n");
				while (r_prev_state && r_prev_state->memory != NULL) {
					//printf ("memory value ");
					if (r_prev_state->memory->location->val) {
						//printf ("%d\n", r_prev_state->memory->location->val->value.int_val);
					}
					else {
						//printf ("null \n");
					}
					r_prev_state->memory = r_prev_state->memory->next;
				}
				while (r_prev_state && r_prev_state->heap != NULL) {
					//printf ("heap value ");
					if (r_prev_state->heap->location->val) {
						//printf ("%d \n", r_prev_state->heap->location->val->value.int_val);
					}
					else {
						//printf ("null \n");
					}
					r_prev_state->heap = r_prev_state->heap->next;
				}
				next_cfg_node = NULL;
			}
		} else {
			//printf ("other type of statement\n");
			check_rule();
			next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
			if (found_second == 1) {
				printf ("property is satisfied\n");
				exit(0);
			}
		}
		curr_cfg_node = next_cfg_node;
	}
	/* printing all the varibales in memory */
	while (r_curr_state && r_curr_state->memory != NULL) {
		r_curr_state->memory = r_curr_state->memory->next;
	}
	//printf ("**************     *******************\n");
	printf ("*************execute finished*********\n");
}

int check_rule () {
	struct cfg_node * temp_cfg_node;
	int node_number;

	temp_cfg_node = r_curr_state->node ;
	node_number = temp_cfg_node->node_number;
	//printf ("node %d %s\n", node_number, r_curr_state->proc->proc_name);
	if (strcmp (r_curr_state->proc->proc_name, "alt_sep_test") == 0 && (node_number == FIRST_NODE_NUMBER)) {
		found_first = 1;
	}
	if ((found_first == 1) && strcmp (r_curr_state->proc->proc_name, "alt_sep_test") == 0 && (node_number == SECOND_NODE_NUMBER)) {
		found_second = 1;
	}
}

void copy_record (s_state * curr_state, s_state * next_state, tree var, int offset) {
	tree type;
	tree fields;
	s_mem_loc_list  lhs_mem;
	s_mem_loc_list  rhs_mem;
	s_value * r_value;
	int k;
	
	//printf ("copy_record\n");
	lhs_mem = find_lvalue (next_state, var);		
	rhs_mem = find_lvalue (curr_state, var);
	for (k = 0; k<offset; k++) {
		lhs_mem = lhs_mem->next;
		rhs_mem = rhs_mem->next;
	}
	r_value = (s_value *) xmalloc (sizeof (s_value));
	if (rhs_mem->location->val) {
		r_value->value = rhs_mem->location->val->value;
		//printf ("record value %d\n", rhs_mem->location->val->value.int_val);
		r_value->kind = rhs_mem->location->val->kind;
	} else {
		r_value = NULL;
	}
	lhs_mem->location->val = r_value;	
	if (r_value && r_value->kind == HEAP_KIND)
		lhs_mem->location->points_to_index = r_value->value.heap_val->location->index;
	type = TREE_TYPE (var);		
	fields = TYPE_FIELDS (type);	
	while (fields) {
		offset = offset + 1;
		if (TREE_CODE (TREE_TYPE (fields)) != RECORD_TYPE) {
			lhs_mem = find_lvalue (next_state, var);		
			rhs_mem = find_lvalue (curr_state, var);
			for (k = 0; k<offset; k++) {
				lhs_mem = lhs_mem->next;
				rhs_mem = rhs_mem->next;
			}
			r_value = (s_value *) xmalloc (sizeof (s_value));
			if (rhs_mem->location->val) {
				r_value->value = rhs_mem->location->val->value;
				//printf ("record value %d\n", rhs_mem->location->val->value.int_val);
				r_value->kind = rhs_mem->location->val->kind;
			} else {
				r_value = NULL;
			}
			lhs_mem->location->val = r_value;	
			if (r_value && r_value->kind == HEAP_KIND)
				lhs_mem->location->points_to_index = r_value->value.heap_val->location->index;
		} else {
			copy_record (curr_state, next_state, var, offset);
		}
		fields = TREE_CHAIN (fields);
	}
	return;
}

s_process * create_process () {
	s_process * new_process;
	
	new_process = (s_process *) xmalloc (sizeof (s_process));
	allocate_curr_state ();
	//printf ("*****child process created******\n");
	return new_process;
}

void copy_shared_vars_among_processes (s_state * curr_state, s_state * next_state) {
	s_mem_loc_list curr_memory;
	s_mem_loc_list temp_memory;
	s_mem_loc_list curr_heap;
	s_mem_loc_list temp_heap;
	
	curr_memory = curr_state->memory;	
 	while (curr_memory != NULL) {
		s_mem_loc * curr_location;
		s_mem_loc * temp_location;
		char * temp_var_name;
		char * curr_var_name;

		curr_location = curr_memory->location;
		curr_var_name = IDENTIFIER_POINTER (DECL_NAME (curr_location->var)); 
		if ((strcmp(curr_var_name, "I_to_A") == 0)
			|| (strcmp(curr_var_name, "A_to_I") == 0)
			|| (strcmp(curr_var_name, "R_to_A") == 0)
			|| (strcmp(curr_var_name, "A_to_R") == 0)) {
			//printf ("curr var %s\n", curr_var_name);
			temp_memory = next_state->memory;
			while (temp_memory != NULL) {
				temp_location = temp_memory->location;
				temp_var_name = IDENTIFIER_POINTER (DECL_NAME (temp_location->var)); 
				if (((strcmp(temp_var_name, "I_to_A") == 0) 
					|| (strcmp(temp_var_name, "A_to_I") == 0)
					|| (strcmp(temp_var_name, "R_to_A") == 0)
					|| (strcmp(temp_var_name, "A_to_R") == 0)) 
					&& (strcmp (curr_var_name, temp_var_name) == 0)){
					/*copy null*/
					//printf ("temp var %s\n", temp_var_name);
					curr_heap = curr_location->val->value.heap_val;
					temp_heap = temp_location->val->value.heap_val;
					/*copy 0*/
					curr_heap = curr_heap->next;
					temp_heap = temp_heap->next;
					temp_heap->location->val->value.int_val = curr_heap->location->val->value.int_val;
					/*copy null*/
					curr_heap = curr_heap->next;
					temp_heap = temp_heap->next;
					/*copy 0*/
					curr_heap = curr_heap->next;
					temp_heap = temp_heap->next;
					temp_heap->location->val->value.int_val = curr_heap->location->val->value.int_val;
					/*copy 0*/
					curr_heap = curr_heap->next;
					temp_heap = temp_heap->next;
					temp_heap->location->val->value.int_val = curr_heap->location->val->value.int_val;
					/*copy null*/
					curr_heap = curr_heap->next;
					temp_heap = temp_heap->next;
					/*copy 0*/
					curr_heap = curr_heap->next;
					temp_heap = temp_heap->next;
					temp_heap->location->val->value.int_val = curr_heap->location->val->value.int_val;
					/*copy 0*/
					curr_heap = curr_heap->next;
					temp_heap = temp_heap->next;
					temp_heap->location->val->value.int_val = curr_heap->location->val->value.int_val;
					/*copy 0*/
					curr_heap = curr_heap->next;
					temp_heap = temp_heap->next;
					temp_heap->location->val->value.int_val = curr_heap->location->val->value.int_val;
					
				}
				temp_memory = temp_memory->next;
			}
		}
		if ((strcmp(curr_var_name, "I_to_A_filled") == 0)
			|| (strcmp(curr_var_name, "A_to_I_filled") == 0)
			|| (strcmp(curr_var_name, "R_to_A_filled") == 0)
			|| (strcmp(curr_var_name, "A_to_R_filled") == 0)) {

			temp_memory = next_state->memory;
			while (temp_memory != NULL) {
				temp_location = temp_memory->location;
				temp_var_name = IDENTIFIER_POINTER (DECL_NAME (temp_location->var)); 
				if (((strcmp(temp_var_name, "I_to_A_filled") == 0) 
					|| (strcmp(temp_var_name, "A_to_I_filled") == 0)
					|| (strcmp(temp_var_name, "R_to_A_filled") == 0)
					|| (strcmp(temp_var_name, "A_to_R_filled") == 0)) 
					&& (strcmp (curr_var_name, temp_var_name) == 0)){
					//printf ("temp var %s %d\n", temp_var_name, curr_location->val->value.int_val);

					temp_location->val->value.int_val = curr_location->val->value.int_val;
				}
				temp_memory = temp_memory->next;
			}
		}
		curr_memory = curr_memory->next;
	}	

}

int * generate_hash_table_entry (s_state * curr_state, int curr_proc_number, int curr_insn_number, 
			s_state * other_state, int other_proc_number, int other_insn_number) {
	s_mem_loc_list temp_memory;
	s_mem_loc_list temp_heap;
	int * hash_table_entry;
	int i = 0;
	
	hash_table_entry = (int *) xmalloc (sizeof (int) * 1000);
	for (i=0; i<400;i++) {
		hash_table_entry[i] = -1;
	}
	i=0;
	hash_table_entry[i] = curr_proc_number;
	i++;
	hash_table_entry[i] = curr_insn_number;
	i++;
	hash_table_entry[i] = other_proc_number;
	i++;
	hash_table_entry[i] = other_insn_number;
	i++;
	temp_heap = curr_state->heap;	
	while (temp_heap != NULL) {
		if (temp_heap->location->val != NULL) {
			if (temp_heap->location->val->kind == INT_KIND) {
				hash_table_entry[i] = temp_heap->location->val->value.int_val;
				i++;
								
			}
		}
		temp_heap = temp_heap->next;
	}
	temp_memory = curr_state->memory;	
	while (temp_memory != NULL) {
		if (temp_memory->location->val != NULL) {
			if (temp_memory->location->val->kind == INT_KIND) {
				hash_table_entry[i] = temp_memory->location->val->value.int_val;
				i++;
			}
		}
		temp_memory = temp_memory->next;
	}


	temp_heap = NULL;
	temp_memory = NULL;
	if (other_state != NULL)
		temp_heap = other_state->heap;	
	while (temp_heap != NULL) {
		if (temp_heap->location->val != NULL) {
			if (temp_heap->location->val->kind == INT_KIND) {
				hash_table_entry[i] = temp_heap->location->val->value.int_val;
				i++;
								
			}
		}
		temp_heap = temp_heap->next;
	}
	if (other_state != NULL)
		temp_memory = other_state->memory;	
	while (temp_memory != NULL) {
		if (temp_memory->location->val != NULL) {
			if (temp_memory->location->val->kind == INT_KIND) {
				hash_table_entry[i] = temp_memory->location->val->value.int_val;
				i++;
			}
		}
		temp_memory = temp_memory->next;
	}

	for (i=0; i<400;i++) {
		if (hash_table_entry[i] != -1 && hash_table_entry[i] != 0 ) {
		//	printf ("%d ", hash_table_entry[i]);
		}
	}
	//printf ("****************************************************\n");
	return hash_table_entry;
}
