
/*========================================================================
               Copyright (C) 2004 by Xiaowan Huang 
                            All rights reserved

    Permission is hereby granted, without written agreement and without
    license or royalty fees, to use, reproduce, prepare derivative
    works, distribute, and display this software and its documentation
    for any purpose, provided that (1) the above copyright notice and
    the following two paragraphs appear in all copies of the source code
    and (2) redistributions, including without limitation binaries,
    reproduce these notices in the supporting documentation. Substantial
    modifications to this software may be copyrighted by their authors
    and need not follow the licensing terms described here, provided
    that the new terms are clearly indicated in all files where they apply.
=======================================================================*/

/*************************************************************************
  FILE:  bitstr.h
  DESCR: 
  AUTH:  Xiaowan Huang
  EMAIL: xhuang@cs.sunysb.edu
  DATE:  (C) Tue Nov 23 23:49:06 2004
*************************************************************************/


#ifndef _BITSTR_H_
#define _BITSTR_H_

#include <limits.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef unsigned char byte_t;

typedef struct bitstr_t {
	byte_t *ptr;
	unsigned int size;	/* in bytes */
	unsigned int nbits;	/* nbits < size*8 */
} bitstr_t;

#define ALLOC(type, len)  		(type *) malloc(sizeof(type) *len)
#define FREE(ptr) 			free(ptr)
#define bitstr_n(bs)			((bs)->nbits)
#define bitstr_size(bs)			((((bs)->nbits) - 1)/CHAR_BIT + 1)
//#define bitstr_elem(bs, bit)	((bs)->ptr)[((bs)->nbits)/CHAR_BIT]
#define bitstr_elem(bs, bit)	((bs)->ptr)[(bit)/CHAR_BIT]
#define bitstr_mask(bs, bit)	(0x80 >> ((bit) % CHAR_BIT))

#define bitstr_isset(bs, bit)	((bitstr_elem(bs, bit) & bitstr_mask(bs, bit)) != 0)
#define bitstr_set(bs, bit) 	((bitstr_elem(bs, bit) |= bitstr_mask(bs, bit)))
#define bitstr_unset(bs, bit)	((bitstr_elem(bs, bit) &= ~bitstr_mask(bs, bit))
#define bitstr_toggle(bs, bit)	((bitstr_elem(bs, bit) ^= bitstr_mask(bs, bit))

bitstr_t *bitstr_alloc(int nbits);
void bitstr_free(bitstr_t *bstr);
bitstr_t *bitstr_dup(const bitstr_t *bstr);
void bitstr_expand(bitstr_t *bstr, int nbits);
void bitstr_append(bitstr_t *bstr, const bitstr_t *bstr2);
void bitstr_append_uint32(bitstr_t *bstr, unsigned long num);
bitstr_t *bitstr_concat(const bitstr_t *bstr1, const bitstr_t *bstr2);
void bitstr_pack(bitstr_t *bstr);
void bitstr_abort(bitstr_t *bstr, int bit);
bitstr_t *bitstr_substr(const bitstr_t *bstr, int from, int to);
unsigned long bitstr_to_uint32(const bitstr_t *bstr);
bitstr_t *bitstr_from_uint32(unsigned long num);
bitstr_t *bitstr_from_uint16(unsigned short num);
unsigned long bitstr_substr_to_uint32(const bitstr_t *bstr, int from, int to);
const char *bitstr_to_bit_seq(const bitstr_t *bstr);
int	bitstr_bit_at(const bitstr_t *bstr, int at);
//void bitstr_printf(const bitstr_t *bstr);

#ifdef __cplusplus
}
#endif

#endif /* _BITSTR_H_ */

