//#include "memwatch/memwatch.h"
//#include "memwatch/memwatch.c"
#include "memory.h"
#include "location.h"
#include "state.h"
#include "globals.h"
#include "process.h"
#include "hashtable/pbl.c"
#include "hashtable/pblhash.c"
#include "hashtable/pbl.h"
#include "execution.h"
#include "macros.h"
#include <time.h>
#include <limits.h>
#include "clone.c"
#include "allocate.c"
#include "lvalue.c"
#include "rvalue.c"
#include "lasso_memoize.h"
#include "value-structures.c"
#include "N.h"

#define RUN_PHIL

#ifdef RUN_TCAS
#include "tcas_macros.h"
#endif

#define MEMWATCH
#define MW_STDIO

int rround = 0;
int runs = 0;
int ** hash_table_key_array;
int removed_keys;
void start_execution(void) 
{
	int i;
	int N;
	void * data;
	struct proc_cfg * main_proc; 
	struct global_list * temp_globals_list;
	char * curr_var_name;
	time_t before, after;
	struct tm * beginTime, * endTime;
 	struct timeval tv;
	struct timezone tz;

	if (gettimeofday (&tv, &tz) < 0) {
		sprintf (stderr, "error in getting time of day\n");
		exit(0);
	}
	srand48(tv.tv_usec);
	
	state_hash_table = pblHtCreate ();
	if (state_hash_table == NULL)
	{
		sprintf (stderr, "error in allocating state_hash_table\n");
		exit(-1);
	}
	hash_table_entry = (int *) xmalloc (sizeof (int) * (MAX_STATE_SIZE * MAX_NUM_PROCESS));
	
	temp_globals_list = globals_list;
 	while (temp_globals_list != NULL) {
		curr_var_name = IDENTIFIER_POINTER (DECL_NAME (temp_globals_list->global)); 
		if (strncmp (curr_var_name, "shared_channel", 14) == 0) {
			shared_variabels++;
		}
		temp_globals_list = temp_globals_list->next;
	}

	N = calculate_N(epsilon_g, delta_g);
	printf ("N = %d\n", N);
	time(&before);
	beginTime= gmtime (&before);
	printf ("beginTime = %d hr %d min %d sec\n", beginTime->tm_hour, beginTime->tm_min, beginTime->tm_sec);
	while (runs < N) {
		int present = 0;

		mm_memory = 0;
		mf_memory = 0;
		length_of_lasso = 0;

		for (i=0; i<MAX_NUM_PROCESS; i++) {
			process_array[i] = NULL;
		}
		/*printf ("********main process ******\n");*/
		main_process = (s_process *) xmalloc (sizeof (s_process));
		mm_memory = mm_memory + sizeof (s_process);
		current_process = NULL;
		num_processes = 0;
		child_process = NULL;
		last_allocated_heap = NULL;
		last_allocated_memory = NULL;
		r_curr_state = NULL;	
		curr_cfg_node = NULL;
		curr_proc = NULL;
		removed_keys = 0;
		state_stack = NULL;
		rround = 0;
		/*identify the main procedure in the list of all procedures*/
		main_proc = procs;
		while (1)
		{
			if (main_proc == NULL ||strcmp (main_proc->proc_name, "main") == 0 )
				break;
			main_proc = main_proc->next;
		}
		if (main_proc == NULL)
		{
			fprintf (stderr, "function <main> is not defined\n");
			exit(0);
		}
		curr_proc = main_proc;
		allocate_curr_state ();	
		main_process->process_state = r_curr_state;
		main_process->process_number = num_processes;
		main_process->curr_proc = curr_proc;
		main_process->curr_cfg_node = curr_cfg_node;
		main_process->last_allocated_memory = last_allocated_memory;
		main_process->last_allocated_heap = last_allocated_heap;
		process_array[num_processes] = main_process;
		num_processes++;
		current_process = main_process;
		curr_cfg_node = curr_proc->cfg_array[0];
		/*start the current execution*/
		first_procedure_allocate_state = 1;
		found_first = 0;
		found_second = 0;

		execute();

		printf ("round = %d  length = %d\n", runs, length_of_lasso);
		for(data = pblHtFirst(state_hash_table ); data; data = pblHtNext(state_hash_table )) {
        		if (pblHtRemove(state_hash_table, 0, 0 ) == 0) {
				/*printf ("Removed an entry\n");*/
				removed_keys++;
			}
    		}		
		//printf ("Removing states\n");
		while (state_stack != NULL) {
			struct state_list * temp_state_list = NULL;
			temp_state_list = pop_state_from_stack ();
			free_state (temp_state_list->state);
			free (temp_state_list->state);
			mf_memory = mf_memory + sizeof (s_state);
			free (temp_state_list);
			mf_memory = mf_memory + sizeof (struct state_list);
		}

		for (i=0; i<MAX_NUM_PROCESS; i++) {
			if (process_array[i]) {
				free (process_array[i]);
				mf_memory = mf_memory + sizeof (s_process);
				process_array[i] = NULL;
			}
		}
		//printf ("mm_memory %d, mf_memory %d*******************\n", mm_memory, mf_memory);
		runs++;	
	}
	time(&after);
	endTime = gmtime (&after);
	printf ("Number of lassos detected = %u\n", number_of_lasso_runs);
	printf ("Number of counter examples = %u\n", N_counter_examples);
	printf ("endTime = %d hr %d min %d sec\n", endTime->tm_hour, endTime->tm_min, endTime->tm_sec);
}

void execute ()
{
	tree curr_insn;
	struct cfg_node * next_cfg_node = NULL;
	s_state * r_prev_state = NULL;
	s_state * r_next_state = NULL;
	s_state * r_other_state = NULL;
	s_mem_loc_list  lhs_mem;

	while (curr_cfg_node != NULL) {
		curr_insn = curr_cfg_node->insn;
		if (curr_insn!= NULL && TREE_CODE (curr_insn) == FUNCTION_DECL) {
			next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
		} else if (curr_insn!= NULL && TREE_CODE (curr_insn) == MODIFY_EXPR) {
			s_value * rhs_value;
			//printf ("modify statements\n");
			if (TREE_CODE (TREE_OPERAND (curr_insn, 1)) == CALL_EXPR) {
				tree formal_params;
				tree actual_params;
				s_value * param_value;
				s_value * global_value;
				char * proc_name;
				struct global_list *temp_globals_list;
				int prev_proc_number, prev_insn_number;

				prev_proc_number = curr_proc->proc_number; 
				prev_insn_number = curr_cfg_node->node_number;
				proc_name = IDENTIFIER_POINTER (DECL_NAME (TREE_OPERAND (TREE_OPERAND (TREE_OPERAND (curr_insn,1), 0), 0))); /*FIXME:*/
				//printf ("** %s ** call expression\n", proc_name);
				if (strcmp (proc_name, "malloc") != 0 && strcmp (proc_name, "VS_toss") != 0 && strcmp (proc_name, "fork") != 0) {
					tree temp_insn;

					r_curr_state->stmt = curr_insn;
					r_curr_state->node = curr_cfg_node;

					temp_insn = TREE_OPERAND (curr_insn, 1);			
					r_prev_state = r_curr_state;
					curr_proc = procs;
					while (1)
					{
						if (curr_proc == NULL ||strcmp (curr_proc->proc_name, proc_name) == 0 )
							break;
						curr_proc = curr_proc->next;
					}
					if (curr_proc == NULL)
					{
						fprintf (stderr, "function <%s> is not defined\n", proc_name);
						exit(0);
					}
					r_curr_state = (s_state *) xmalloc (sizeof (s_state)); 
					mm_memory = mm_memory + sizeof (s_state);
					r_curr_state->return_state = r_prev_state;
					r_curr_state->return_stack[r_prev_state->return_stack_length + 1] = prev_proc_number;
					r_curr_state->return_stack[r_prev_state->return_stack_length + 2] = prev_insn_number; 
					r_curr_state->return_stack_length = r_prev_state->return_stack_length + 2;
					clone_memory_heap_while_function_call_or_return (r_prev_state, r_curr_state);
					formal_params = DECL_ARGUMENTS (curr_proc->proc);
					actual_params = TREE_OPERAND (temp_insn, 1);
					while (formal_params && actual_params) {
						lhs_mem = find_lvalue (r_curr_state, formal_params);
						param_value = find_rvalue (r_prev_state, TREE_VALUE (actual_params));
						if (param_value && param_value->kind == HEAP_KIND) {
							lhs_mem->location->points_to_index = param_value->value.heap_val->location->index;
							free (param_value);
							mf_memory = mf_memory + sizeof (s_value);
							param_value = NULL;
						} else {
							if (lhs_mem->location->val != NULL) {
								free (lhs_mem->location->val);
								mf_memory = mf_memory + sizeof (s_value);
								lhs_mem->location->val = NULL;
							}
							lhs_mem->location->val = param_value;
						}
						formal_params = TREE_CHAIN (formal_params);
						actual_params = TREE_CHAIN (actual_params);
					}
					current_process->curr_proc = curr_proc;
					current_process->curr_cfg_node = curr_cfg_node;
					next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
				} else if (strcmp (proc_name, "VS_toss") != 0 && strcmp (proc_name, "fork") != 0) {
					int malloc_size;
					tree type;
					tree temp_insn;

					r_curr_state->stmt = curr_insn;
					r_curr_state->node = curr_cfg_node;
					current_process->curr_proc = curr_proc;
					current_process->curr_cfg_node = curr_cfg_node;

					temp_insn = TREE_OPERAND (curr_insn, 1);
					lhs_mem = find_lvalue (r_curr_state, TREE_OPERAND (curr_insn, 0));
					malloc_size = TREE_VALUE (TREE_OPERAND (temp_insn, 1))->int_cst.int_cst.low ;
					type = TREE_TYPE (curr_insn);
					rhs_value = create_value_structure_for_pointer (malloc_size, type);
					if (lhs_mem != NULL) {
				
						if (TREE_CODE (TREE_OPERAND (curr_insn, 0)) == ARRAY_REF) {
							lhs_mem->location->val = rhs_value;
							lhs_mem->location->points_to_index = rhs_value->value.heap_val->location->index;
						} else if (TREE_CODE (TREE_OPERAND (curr_insn, 0)) == INDIRECT_REF) {
							lhs_mem->location->val = rhs_value;
							lhs_mem->location->points_to_index = rhs_value->value.heap_val->location->index;
						} else {
							lhs_mem->location->val = rhs_value;
							lhs_mem->location->points_to_index = rhs_value->value.heap_val->location->index;
						}
					}
					/*CHECK_HASH_TABLE();*/
					next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
				} else if (strcmp (proc_name, "fork") != 0) {
					tree temp_insn;
					int range;

					r_curr_state->stmt = curr_insn;
					r_curr_state->node = curr_cfg_node;
					current_process->curr_proc = curr_proc;
					current_process->curr_cfg_node = curr_cfg_node;

					temp_insn = TREE_OPERAND (curr_insn, 1);
					lhs_mem = find_lvalue (r_curr_state, TREE_OPERAND (curr_insn, 0));
					rhs_value = find_rvalue (r_curr_state, TREE_VALUE (TREE_OPERAND (temp_insn, 1)));
					if (rhs_value != NULL) {
						range = rhs_value->value.int_val;
					} else {
						range = -1;
					}

					if (rhs_value != NULL) {
						free (rhs_value);
						mf_memory = mf_memory + sizeof (s_value);
						rhs_value = NULL;
					}

					rhs_value = (s_value *) xmalloc (sizeof (s_value));
					mm_memory = mm_memory + sizeof (s_value);
					/*rhs_value->value.int_val = ((range+1) * (rand()/(RAND_MAX + 1.0)));*/
					double drand = drand48();
					rhs_value->value.int_val = (int)(drand * (double) (range+1));
					//printf ("VS_toss(%d) = %d\n",range,  rhs_value->value.int_val);
					rround++;
					if (lhs_mem != NULL) {
						
						if (lhs_mem->location->val != NULL) {
							free (lhs_mem->location->val);
							mf_memory = mf_memory + sizeof (s_value);
							lhs_mem->location->val = NULL;
						}
						
						if (TREE_CODE (TREE_OPERAND (curr_insn, 0)) == ARRAY_REF) {
							lhs_mem->location->val = rhs_value;
						} else if (TREE_CODE (TREE_OPERAND (curr_insn, 0)) == INDIRECT_REF) {
							lhs_mem->location->val = rhs_value;
						} else {
							lhs_mem->location->val = rhs_value;
						}
					}
					/*CHECK_HASH_TABLE();*/
					next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
				} else {
					tree temp_insn;
					struct cfg_node * prev_cfg_node;
					s_mem_loc_list prev_last_allocated_memory;
					s_mem_loc_list prev_last_allocated_heap;

					r_curr_state->stmt = curr_insn;
					r_curr_state->node = curr_cfg_node;
					current_process->curr_proc = curr_proc;
					current_process->curr_cfg_node = curr_cfg_node;

					temp_insn = TREE_OPERAND (curr_insn, 1);
					lhs_mem = find_lvalue (r_curr_state, TREE_OPERAND (curr_insn, 0));
					rhs_value = (s_value *) xmalloc (sizeof (s_value));
					mm_memory = mm_memory + sizeof (s_value);
					rhs_value->value.int_val = 0;
					if (lhs_mem != NULL) {
						
						if (lhs_mem->location->val) {
							free (lhs_mem->location->val);
							mf_memory = mf_memory + sizeof (s_value);
							lhs_mem->location->val = NULL;
						}
						

						if (TREE_CODE (TREE_OPERAND (curr_insn, 0)) == ARRAY_REF) {
							lhs_mem->location->val = rhs_value;
						} else if (TREE_CODE (TREE_OPERAND (curr_insn, 0)) == INDIRECT_REF) {
							lhs_mem->location->val = rhs_value;
						} else {
							lhs_mem->location->val = rhs_value;
						}
					}
					next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
					curr_cfg_node = next_cfg_node;
					if (curr_cfg_node != NULL) {
						curr_insn = curr_cfg_node->insn;
					}
					r_prev_state = r_curr_state;
					prev_cfg_node = curr_cfg_node;
					prev_last_allocated_memory = last_allocated_memory;
					prev_last_allocated_heap = last_allocated_heap;

					child_process = create_process();
					next_cfg_node = curr_proc->cfg_array[prev_cfg_node->outgoing_edges->node[1]];
					copy_shared_channels_and_variables(r_prev_state, r_curr_state);
					child_process->process_state =  r_curr_state;
					child_process->process_number = num_processes;
					child_process->process_state->stmt = next_cfg_node->insn;
					child_process->process_state->node = next_cfg_node;
					child_process->curr_proc = curr_proc;
					child_process->curr_cfg_node = next_cfg_node;
					child_process->last_allocated_memory = last_allocated_memory;
					child_process->last_allocated_heap = last_allocated_heap;

					next_cfg_node = curr_proc->cfg_array[prev_cfg_node->outgoing_edges->next->node[1]];
					curr_cfg_node = next_cfg_node;
					current_process->process_state = r_prev_state; 
					current_process->curr_proc = curr_proc;
					current_process->curr_cfg_node = curr_cfg_node;
					current_process->last_allocated_memory = prev_last_allocated_memory;
					current_process->last_allocated_heap = prev_last_allocated_heap;
					r_curr_state = r_prev_state;
					last_allocated_memory = prev_last_allocated_memory;
					last_allocated_heap = prev_last_allocated_heap;
					process_array[num_processes] = child_process;
					num_processes++;
				}
			} else {
				tree temp_insn;	

				temp_insn = TREE_OPERAND (curr_insn, 1);
				r_curr_state->stmt = curr_insn;
				r_curr_state->node = curr_cfg_node;
				current_process->curr_proc = curr_proc;
				current_process->curr_cfg_node = curr_cfg_node;
				lhs_mem = find_lvalue (r_curr_state, TREE_OPERAND (curr_insn, 0));
				if (TREE_CODE (temp_insn) == CONVERT_EXPR) {
					rhs_value = find_rvalue (r_curr_state, TREE_OPERAND (temp_insn, 0));
				} else if (TREE_CODE (temp_insn) == NOP_EXPR) {
					rhs_value = find_rvalue (r_curr_state, TREE_OPERAND (temp_insn, 0));
				} else {
					rhs_value = find_rvalue (r_curr_state, temp_insn);
				}
				if (lhs_mem != NULL) {

					if (lhs_mem->location->val != NULL) {
						free (lhs_mem->location->val);
						mf_memory = mf_memory + sizeof (s_value);
						lhs_mem->location->val = NULL;
					}
					
					if (TREE_CODE (TREE_OPERAND (curr_insn, 0)) == ARRAY_REF) {
						lhs_mem->location->val = rhs_value;
						if (rhs_value->kind == HEAP_KIND)
							lhs_mem->location->points_to_index = rhs_value->value.heap_val->location->index;
					} else if (TREE_CODE (TREE_OPERAND (curr_insn, 0)) == INDIRECT_REF) {
						lhs_mem->location->val = rhs_value;
						if (rhs_value->kind == HEAP_KIND)
							lhs_mem->location->points_to_index = rhs_value->value.heap_val->location->index;
					} else {
						lhs_mem->location->val = rhs_value;
						if (rhs_value && rhs_value->kind == HEAP_KIND)
							lhs_mem->location->points_to_index = rhs_value->value.heap_val->location->index;
						if (rhs_value && rhs_value->kind == MEM_KIND)
							lhs_mem->location->points_to_index = rhs_value->value.mem_val->location->index;
					}
				} else {
					if (rhs_value) {
						free (rhs_value);
						mf_memory = mf_memory + sizeof (s_value);
						rhs_value = NULL;
					}
				}
				/*CHECK_HASH_TABLE();*/
				next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
			}
		} else if (curr_insn!= NULL && TREE_CODE (curr_insn) == COND_EXPR) {
			s_value * cond_value;

			//printf ("conditional statement\n");
			r_curr_state->stmt = curr_insn;
			r_curr_state->node = curr_cfg_node;
			cond_value = find_rvalue (r_curr_state, COND_EXPR_COND (curr_insn));		
			if (cond_value->value.int_val == 1 || cond_value->value.bool_val == 1) {
				next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
			}
			else {
				next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->next->node[1]];
			}
			if (cond_value != NULL) {
				free (cond_value);
				mf_memory = mf_memory + sizeof (s_value);
				cond_value = NULL;
			}
		} else if (curr_insn!= NULL && TREE_CODE (curr_insn) == CALL_EXPR) {
			tree formal_params;
			tree actual_params;
			s_value * param_value;
			s_value * global_value;
			char * proc_name;
			struct global_list *temp_globals_list;
			int prev_proc_number, prev_insn_number;
			prev_proc_number = curr_proc->proc_number; 
			prev_insn_number = curr_cfg_node->node_number;

			r_curr_state->stmt = curr_insn;
			r_curr_state->node = curr_cfg_node;
			current_process->curr_proc = curr_proc;
			current_process->curr_cfg_node = curr_cfg_node;
			r_prev_state = r_curr_state;
			proc_name = IDENTIFIER_POINTER (DECL_NAME (TREE_OPERAND (TREE_OPERAND (curr_insn, 0), 0))); /*FIXME:*/
			//printf ("** %s ** call statement\n", proc_name);
			if (strcmp (proc_name, "switch_process") != 0) { 
				curr_proc = procs;
				while (1)
				{
					if (curr_proc == NULL ||strcmp (curr_proc->proc_name, proc_name) == 0 )
						break;
					curr_proc = curr_proc->next;
				}
				if (curr_proc == NULL)
				{
					fprintf (stderr, "function <%s> is not defined\n", proc_name);
					exit(0);
				}
				//allocate_curr_state ();
				r_curr_state = (s_state *) xmalloc (sizeof (s_state)); 
				mm_memory = mm_memory + sizeof (s_state);
				r_curr_state->return_state = r_prev_state;
				r_curr_state->return_stack[r_prev_state->return_stack_length + 1] = prev_proc_number;
				r_curr_state->return_stack[r_prev_state->return_stack_length + 2] = prev_insn_number; 
				r_curr_state->return_stack_length = r_prev_state->return_stack_length + 2;
			
				clone_memory_heap_while_function_call_or_return (r_prev_state, r_curr_state);
				formal_params = DECL_ARGUMENTS (curr_proc->proc);
				actual_params = TREE_OPERAND (curr_insn, 1);
				while (formal_params && actual_params) {
					lhs_mem = find_lvalue (r_curr_state, formal_params);
					param_value = find_rvalue (r_prev_state, TREE_VALUE (actual_params));
					if (param_value && param_value->kind == HEAP_KIND) {
						lhs_mem->location->points_to_index = param_value->value.heap_val->location->index;
						free (param_value);
						mf_memory = mf_memory + sizeof (s_value);
						param_value = NULL;
					} else {
						if (lhs_mem->location->val != NULL) {
							free (lhs_mem->location->val);
							mf_memory = mf_memory + sizeof (s_value);
							lhs_mem->location->val = NULL;
						}
						lhs_mem->location->val = param_value;
					}
					formal_params = TREE_CHAIN (formal_params);
					actual_params = TREE_CHAIN (actual_params);
				}
				//clone_heap (r_prev_state, r_curr_state);
				//next_cfg_node = curr_proc->cfg_array[1];
				current_process->curr_proc = curr_proc;
				current_process->curr_cfg_node = curr_cfg_node;
				next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
			} else {
				s_state * prev_state;
				/* switch_process */
				//printf ("current process number %d %d \n", current_process->process_number, num_processes);
				CHECK_HASH_TABLE();
				{
					int new_process_number;
					int prev_process_number;
					prev_state = r_curr_state;
					current_process->process_state = r_curr_state;
					current_process->curr_proc = curr_proc;
					next_cfg_node = curr_cfg_node;
					current_process->curr_cfg_node = next_cfg_node;
					current_process->last_allocated_memory = last_allocated_memory;
					current_process->last_allocated_heap = last_allocated_heap;
					prev_process_number = current_process->process_number;
					process_array[current_process->process_number] = current_process;

					double drand = drand48();
					new_process_number = (int)(drand * (double) (num_processes));
					while (new_process_number == prev_process_number) {
						double drand = drand48();
						new_process_number = (int)(drand * (double) (num_processes));
					}
					//printf ("new process number %d \n", new_process_number);
					current_process = process_array[new_process_number];
					curr_cfg_node = current_process->curr_cfg_node;
					curr_proc = current_process->curr_proc;
					r_curr_state = current_process->process_state;
					copy_shared_channels_and_variables(prev_state, r_curr_state);
					last_allocated_memory = current_process->last_allocated_memory;
					last_allocated_heap = current_process->last_allocated_heap;
				}
				next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
			}
		} else if (curr_insn!= NULL && TREE_CODE (curr_insn) == RETURN_EXPR) {
			s_value * return_value;
			s_mem_loc_list temp_memory;
			s_mem_loc_list temp_memorya;

			//printf ("return statement\n");
			if (TREE_OPERAND (curr_insn, 0)) {
				curr_insn = TREE_OPERAND (curr_insn, 0);
			}	
			if (TREE_CODE (curr_insn) == MODIFY_EXPR) {
				return_value = find_rvalue (r_curr_state, TREE_OPERAND (curr_insn, 1));	
			} else {
				return_value = find_rvalue (r_curr_state, curr_insn);	
			}
			/*specially for TCAS*/
			#ifdef RUN_TCAS
			int procedure_name_is_initialize = 0;
			if (strcmp (curr_proc->proc_name, "initialize") == 0) {
				procedure_name_is_initialize = 1;
			}
			#endif
			
			/*For Needham-Schroeder*/
			if (strcmp (curr_proc->proc_name, "VS_assert") == 0) {
				//printf ("return value %d\n", return_value->value.int_val);
				if (return_value->value.int_val == 0) {
					printf ("Assertion is not valid at this point\n");
					prev_N_counter_examples = N_counter_examples;
					N_counter_examples++;
					exit(0); //For Geometric, as we stop after 1st counter example 
					return; //For Quantitative, as we do not stop after 1st counter example 
				}
			}	
			r_prev_state = r_curr_state;
			r_curr_state = r_prev_state->return_state;
			/*if return state is null means this is the <main> procedure 
			i.e. first proc in the call graph */
			if (r_curr_state != NULL) {
				struct global_list *temp_globals_list;
				s_value * global_value;
				s_mem_loc_list lhs_mem;

				r_curr_state->return_stack[r_prev_state->return_stack_length] = -1;	
				r_curr_state->return_stack[r_prev_state->return_stack_length - 1] = -1;
				r_curr_state->return_stack_length = r_prev_state->return_stack_length - 2;
				curr_proc = r_curr_state->proc;
				curr_insn = r_curr_state->stmt;
				curr_cfg_node = r_curr_state->node;

				temp_globals_list = globals_list;
				while (temp_globals_list != NULL) {
					if (TREE_CODE (TREE_TYPE (temp_globals_list->global)) == RECORD_TYPE) {
						copy_record (r_prev_state, r_curr_state, temp_globals_list->global, 0);
					} else {
						lhs_mem = find_lvalue (r_curr_state, temp_globals_list->global);	
						global_value = find_rvalue (r_prev_state, temp_globals_list->global);
						if (global_value && global_value->kind == HEAP_KIND) {
							lhs_mem->location->points_to_index = global_value->value.heap_val->location->index;
							free (global_value);
							mf_memory = mf_memory + sizeof (s_value);
							global_value = NULL;
						} else {
							
							if (lhs_mem->location->val != NULL) {
								free (lhs_mem->location->val);
								mf_memory = mf_memory + sizeof (s_value);
								lhs_mem->location->val = NULL;
							}
							
							lhs_mem->location->val = global_value;
						}
					}
					temp_globals_list = temp_globals_list->next;
				}
				//clone_heap (r_prev_state, r_curr_state);
				{
					s_mem_loc_list temp_heap, prev_heap;
					s_mem_loc_list curr_heap, stored_new_heap;
					s_mem_loc_list curr_heap1, curr_heap2;
					s_mem_loc_list curr_memory1;

					prev_heap = r_prev_state->heap;
					temp_heap = r_curr_state->heap;
					while (prev_heap != NULL && temp_heap != NULL) {
						if (prev_heap->location->val != NULL) {
							if (prev_heap->location->val->kind == INT_KIND) {
								temp_heap->location->val->value.int_val = prev_heap->location->val->value.int_val; 
								temp_heap->location->points_to_index = prev_heap->location->points_to_index; 
								temp_heap->location->val->kind = INT_KIND;
							}
						}
						stored_new_heap = temp_heap;
						temp_heap = temp_heap->next;
						prev_heap = prev_heap->next;
					}
					/* it is assumed that heap of the function where it is 
					returning will be larger than the function from where it
					return*/
					curr_heap = prev_heap;
					while (curr_heap != NULL) {
						s_mem_loc *temp_location;
						s_mem_loc *curr_location;

						temp_heap = (s_mem_loc_list) xmalloc (sizeof (struct my_mem_list));
						mm_memory = mm_memory + sizeof (struct my_mem_list);
						curr_location = curr_heap->location;
						temp_location = (s_mem_loc *) xmalloc (sizeof (s_mem_loc));
						mm_memory = mm_memory + sizeof (s_mem_loc);
						temp_location->val = NULL;
						temp_location->var = curr_location->var; 
						temp_location->next_struct = curr_location->next_struct; 
						temp_location->index = curr_location->index; 
						if (curr_location->val != NULL) {
							temp_location->val = (s_value *) xmalloc (sizeof (s_value));
							mm_memory = mm_memory + sizeof (s_value);
							if (curr_location->val->kind == INT_KIND) {
								temp_location->val->value.int_val = curr_location->val->value.int_val; 
								temp_location->points_to_index = curr_location->points_to_index; 
								temp_location->val->kind = INT_KIND;
							}
							else if (curr_location->val->kind == FLOAT_KIND) {
								temp_location->val->value.float_val = curr_location->val->value.float_val; 
								temp_location->points_to_index = curr_location->points_to_index; 
								temp_location->val->kind = FLOAT_KIND;
							}
							else if (curr_location->val->kind == BOOL_KIND) {
								temp_location->val->value.bool_val = curr_location->val->value.bool_val; 
								temp_location->points_to_index = curr_location->points_to_index; 
								temp_location->val->kind = BOOL_KIND;
							}
							else if (curr_location->val->kind == MEM_KIND) {
								temp_location->val->value.mem_val = curr_location->val->value.mem_val; 
								temp_location->points_to_index = curr_location->points_to_index; 
								temp_location->val->kind = MEM_KIND;
							}
							else if (curr_location->val->kind == HEAP_KIND) {
								temp_location->val->value.heap_val = curr_location->val->value.heap_val; 
								temp_location->points_to_index = curr_location->points_to_index; 
								temp_location->val->kind = HEAP_KIND;
							}
						} else {
							temp_location->val = NULL;
							temp_location->points_to_index = -1; 
						}
						temp_heap->location = temp_location;
						temp_heap->next = NULL;
						if (stored_new_heap != NULL) { 
							stored_new_heap->next = temp_heap;
							stored_new_heap = temp_heap;
						} else {
							r_curr_state->heap = temp_heap;
							stored_new_heap = temp_heap;
						}
						curr_heap = curr_heap->next;	
					}
					/*renew tha value of last allocated heap*/
					temp_heap = r_curr_state->heap; 		
					last_allocated_heap = temp_heap;
					while (temp_heap != NULL) {
						last_allocated_heap = temp_heap;
						temp_heap = temp_heap->next;
					}
					/* copying the links of heap locations */
					/* if a heap location is pointing to another heap location, then find that link and put into first's value */
					curr_heap1 = r_curr_state->heap;
					while (curr_heap1 != NULL) {
						int i;
						if (curr_heap1->location->points_to_index >0){
							curr_heap2 = r_curr_state->heap;
							i = 0;
							while (curr_heap2 != NULL && i!= curr_heap1->location->points_to_index) {
								curr_heap2 = curr_heap2->next;
								i++;
							}
							
							if (i == curr_heap1->location->points_to_index) {
								curr_heap1->location->val->value.heap_val = curr_heap2;
							}
						} 
						curr_heap1 = curr_heap1->next;
					}
					/* copying the links of memory locations */
					/* if a memory location is pointing to another heap location, then find that link and put into first's value */
					//printf ("h1\n");
					curr_memory1 = r_curr_state->memory;
					while (curr_memory1 != NULL) {
						int i;
						if (curr_memory1->location->points_to_index >=0){
							curr_heap2 = r_curr_state->heap;
							i = 0;
							while (curr_heap2 != NULL && i!= curr_memory1->location->points_to_index) {
								curr_heap2 = curr_heap2->next;
								i++;
							}
							
							if (i == curr_memory1->location->points_to_index) {
								curr_memory1->location->val->value.heap_val = curr_heap2;
							}
						} 
						curr_memory1 = curr_memory1->next;
					}
					//printf ("h2\n");
					
					/*copy the next_srtuct field */
					curr_heap1 = r_curr_state->heap;
					while (curr_heap1 != NULL) {
						int i;
						int index;
						if (curr_heap1->location->next_struct){
							i = 0;
							index = curr_heap1->location->next_struct->location->index;
							curr_heap2 = r_curr_state->heap;
							while (curr_heap2 != NULL && i!= index) {
								curr_heap2 = curr_heap2->next;
								i++;
							}
							if (i == index ) {
								curr_heap1->location->next_struct = curr_heap2;
							}
						} 
						curr_heap1 = curr_heap1->next;
					}
				}
				/*specially for TCAS*/
				#ifdef RUN_TCAS
				if (procedure_name_is_initialize == 1) {
					s_mem_loc_list temp_memory;
					temp_memory = r_curr_state->memory;	
					while (temp_memory != NULL) {
						if (temp_memory->location->var != NULL) {
							if (strcmp (IDENTIFIER_POINTER (DECL_NAME (temp_memory->location->var)), "Up_Separation") == 0) {
								up_sep = temp_memory->location->val->value.int_val;
							}
							if (strcmp (IDENTIFIER_POINTER (DECL_NAME (temp_memory->location->var)), "Down_Separation") == 0) {
								down_sep = temp_memory->location->val->value.int_val;
							}
							if (strcmp (IDENTIFIER_POINTER (DECL_NAME (temp_memory->location->var)), "Own_Tracked_Alt") == 0) {
								own_tracked_alt = temp_memory->location->val->value.int_val;
							}
							if (strcmp (IDENTIFIER_POINTER (DECL_NAME (temp_memory->location->var)), "Other_Tracked_Alt") == 0) {
								other_tracked_alt = temp_memory->location->val->value.int_val;
							}
							if (strcmp (IDENTIFIER_POINTER (DECL_NAME (temp_memory->location->var)), "Alt_Layer_Value") == 0) {
								alt_layer_value = temp_memory->location->val->value.int_val;
							}
							if (alt_layer_value == 0 && strcmp (IDENTIFIER_POINTER (DECL_NAME (temp_memory->location->var)), "Positive_RA_Alt_Thresh0") == 0) {
								positive_ra_alt_thresh = temp_memory->location->val->value.int_val;
							}
							if (alt_layer_value == 1 && strcmp (IDENTIFIER_POINTER (DECL_NAME (temp_memory->location->var)), "Positive_RA_Alt_Thresh1") == 0) {
								positive_ra_alt_thresh = temp_memory->location->val->value.int_val;
							}
							if (alt_layer_value == 2 && strcmp (IDENTIFIER_POINTER (DECL_NAME (temp_memory->location->var)), "Positive_RA_Alt_Thresh2") == 0) {
								positive_ra_alt_thresh = temp_memory->location->val->value.int_val;
							}
							if (alt_layer_value == 3 && strcmp (IDENTIFIER_POINTER (DECL_NAME (temp_memory->location->var)), "Positive_RA_Alt_Thresh3") == 0) {
								positive_ra_alt_thresh = temp_memory->location->val->value.int_val;
							}
						}
						temp_memory = temp_memory->next;
					}
					if (check_P2 () == 0)
						return;
				}
				#endif
				/*specially for TCAS ends*/

				/* if the target of return is assignment then update the 
				value of lhs by return value*/
				if (TREE_CODE (curr_insn) == MODIFY_EXPR) {
					lhs_mem = find_lvalue (r_curr_state, TREE_OPERAND (curr_insn, 0));
					if (lhs_mem != NULL) {

						
						if (lhs_mem->location->val != NULL) {
							free (lhs_mem->location->val);
							mf_memory = mf_memory + sizeof (s_value);
							lhs_mem->location->val = NULL;
						}
						

						if (TREE_CODE (TREE_OPERAND (curr_insn, 0)) == ARRAY_REF) {
							lhs_mem->location->val = return_value;
							if (return_value->kind == HEAP_KIND)
								lhs_mem->location->points_to_index = return_value->value.heap_val->location->index;
						} else if (TREE_CODE (TREE_OPERAND (curr_insn, 0)) == INDIRECT_REF) {
							lhs_mem->location->val = return_value;
							if (return_value->kind == HEAP_KIND)
								lhs_mem->location->points_to_index = return_value->value.heap_val->location->index;
						} else {
							lhs_mem->location->val = return_value;
							if (return_value->kind == HEAP_KIND)
								lhs_mem->location->points_to_index = return_value->value.heap_val->location->index;
						}
					} else {
						if (return_value) {
							free (return_value);
							mf_memory = mf_memory + sizeof (s_value);
							return_value = NULL;
						}
					}
				}
				/*CHECK_HASH_TABLE();*/
				next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
			} else {
				//printf ("************** values *******************\n");
				while (r_prev_state && r_prev_state->memory != NULL) {
					//printf ("memory value ");
					if (r_prev_state->memory->location->val) {
						//printf ("%d\n", r_prev_state->memory->location->val->value.int_val);
					}
					else {
						//printf ("null \n");
					}
					r_prev_state->memory = r_prev_state->memory->next;
				}
				while (r_prev_state && r_prev_state->heap != NULL) {
					//printf ("heap value ");
					if (r_prev_state->heap->location->val) {
						//printf ("%d \n", r_prev_state->heap->location->val->value.int_val);
					}
					else {
						//printf ("null \n");
					}
					r_prev_state->heap = r_prev_state->heap->next;
				}
				curr_proc = NULL;
				curr_cfg_node = NULL;
				next_cfg_node = NULL;
				return;
			}
			current_process->process_state = r_curr_state;
			current_process->curr_proc = curr_proc;
			current_process->curr_cfg_node = curr_cfg_node;
			free_state (r_prev_state);
		} else {
			//printf ("other type of statement\n");
			r_curr_state->stmt = curr_insn;
			r_curr_state->node = curr_cfg_node;
			next_cfg_node = curr_proc->cfg_array[curr_cfg_node->outgoing_edges->node[1]];
		}
		/*For TCAS*/
		#ifdef RUN_TCAS
		/*{	
			check_rule();
			if (found_second == 1) {
				printf ("property is satisfied\n");
				prev_N_counter_examples = N_counter_examples;
				N_counter_examples++;
				exit(0); //For Geometric, as we stop after 1st counter example 
				return; //For Quantitative, as we do not stop after 1st counter example 
			}
			
		}*/
		#endif
		curr_cfg_node = next_cfg_node;
	}
}

void copy_record (s_state * curr_state, s_state * next_state, tree var, int offset) {
	s_mem_loc_list  lhs_mem;
	s_mem_loc_list  rhs_mem;
	s_value * r_value;
	tree fields;
	tree type;
	int k;
	
	//printf ("copy_record\n");
	lhs_mem = find_lvalue (next_state, var);		
	rhs_mem = find_lvalue (curr_state, var);
	for (k = 0; k<offset; k++) {
		lhs_mem = lhs_mem->next;
		rhs_mem = rhs_mem->next;
	}
	r_value = (s_value *) xmalloc (sizeof (s_value));
	mm_memory = mm_memory + sizeof (s_value);
	if (rhs_mem->location->val) {
		r_value->value = rhs_mem->location->val->value;
		r_value->kind = rhs_mem->location->val->kind;
		//printf ("record value %d\n", rhs_mem->location->val->value.int_val);
	} else {
		//printf ("record value null\n");
		free (r_value);
		mf_memory = mf_memory + sizeof (s_value);
		r_value = NULL;
	}
	
	if (lhs_mem->location->val) {
		free (lhs_mem->location->val);
		mf_memory = mf_memory + sizeof (s_value);
		lhs_mem->location->val = NULL;
	}
	

	lhs_mem->location->val = r_value;	
	if (r_value && r_value->kind == HEAP_KIND)
		lhs_mem->location->points_to_index = r_value->value.heap_val->location->index;
	type = TREE_TYPE (var);		
	fields = TYPE_FIELDS (type);	
	while (fields) {
		offset = offset + 1;
		if (TREE_CODE (TREE_TYPE (fields)) != RECORD_TYPE) {
			lhs_mem = find_lvalue (next_state, var);		
			rhs_mem = find_lvalue (curr_state, var);
			for (k = 0; k<offset; k++) {
				lhs_mem = lhs_mem->next;
				rhs_mem = rhs_mem->next;
			}
			r_value = (s_value *) xmalloc (sizeof (s_value));
			mm_memory = mm_memory + sizeof (s_value);
			if (rhs_mem->location->val) {
				r_value->value = rhs_mem->location->val->value;
				r_value->kind = rhs_mem->location->val->kind;
				//printf ("record value %d\n", rhs_mem->location->val->value.int_val);
			} else {
				//printf ("record value null\n");
				free (r_value);
				mf_memory = mf_memory + sizeof (s_value);
				r_value = NULL;
			}
			
			if (lhs_mem->location->val) {
				free (lhs_mem->location->val);
				mf_memory = mf_memory + sizeof (s_value);
				lhs_mem->location->val = NULL;
			}
			
			lhs_mem->location->val = r_value;	
			if (r_value && r_value->kind == HEAP_KIND)
				lhs_mem->location->points_to_index = r_value->value.heap_val->location->index;
		} else {
			copy_record (curr_state, next_state, var, offset);
		}
		fields = TREE_CHAIN (fields);
	}
	return;
}

s_process * create_process () {
	s_process * new_process;
	
	new_process = (s_process *) xmalloc (sizeof (s_process));
	mm_memory = mm_memory + sizeof (s_process);
	allocate_curr_state ();
	return new_process;
}

void copy_records_on_heap (s_mem_loc_list * curr_heap, s_mem_loc_list * next_heap, tree type) {
	s_mem_loc_list temp_curr_heap;
	s_mem_loc_list temp_next_heap;
	tree fields;

	fields = TYPE_FIELDS (type);
	temp_curr_heap = *curr_heap;
	temp_next_heap = *next_heap;
	while (fields) {
		*curr_heap = (*curr_heap)->next;
		*next_heap = (*next_heap)->next;
		if (TREE_CODE (TREE_TYPE (fields)) != RECORD_TYPE) {
			(*next_heap)->location->val->value.int_val = (*curr_heap)->location->val->value.int_val; 	
		}
		else {
			copy_records_on_heap (curr_heap, next_heap, TREE_TYPE (fields));
		}
		fields = TREE_CHAIN (fields);
	}
}

void copy_shared_channels_and_variables (s_state * curr_state, s_state * next_state) {
	s_mem_loc_list curr_memory;
	s_mem_loc_list temp_memory;
	s_mem_loc_list curr_heap;
	s_mem_loc_list temp_heap;
	s_mem_loc * curr_location;
	s_mem_loc * temp_location;
	char * temp_var_name;
	char * curr_var_name;
	int i;
	//printf ("copy_shared_channels_and_variables \n");
	curr_memory = curr_state->memory;	
	while (curr_memory != NULL) {
		curr_location = curr_memory->location;
		if (curr_location->var != NULL ) {
			curr_var_name = IDENTIFIER_POINTER (DECL_NAME (curr_location->var)); 
			if (strncmp (curr_var_name, "shared_channel", 14) == 0) {
				temp_memory = next_state->memory;
				while (temp_memory != NULL) {
					temp_location = temp_memory->location;
					if (temp_location->var != NULL ) {
						temp_var_name = IDENTIFIER_POINTER (DECL_NAME (temp_location->var)); 
						if (strcmp(temp_var_name, curr_var_name) == 0) {
							/* it is assumed that shared channel can be of array type only
							   otherwise shared variable will be of integer type */
							if (TREE_CODE (TREE_TYPE (temp_location->var)) != ARRAY_TYPE) {
								if (curr_location->val && temp_location->val)
									temp_location->val->value.int_val = curr_location->val->value.int_val;
								else 
									temp_location->val = NULL;
							} else {
								/*shared array will have record type elements only nothing else
								and records can not have an array or pointer as their field*/
								int array_size;
								int j;
								tree tmp = TREE_TYPE (temp_location->var);
								if (curr_location->val && temp_location->val) {
									curr_heap = curr_location->val->value.heap_val;
									temp_heap = temp_location->val->value.heap_val;
									array_size = TREE_INT_CST_LOW (TYPE_SIZE (tmp)) /
											TREE_INT_CST_LOW (TYPE_SIZE (TREE_TYPE (tmp)));
									for (j=0; j<array_size; j++) {
										copy_records_on_heap (&curr_heap, &temp_heap, TREE_TYPE (TREE_TYPE (temp_location->var)));
										curr_heap = curr_heap->next; 
										temp_heap = temp_heap->next;
									}
								}
							}
						}
					}
					temp_memory = temp_memory->next;
				}
			}
		}
		curr_memory = curr_memory->next;
	}
	//printf ("finished copy_shared_channels_and_variables \n");
}

void generate_hash_table_entry () {
	s_state * temp_curr_state;
	s_process * temp_current_process;
	s_mem_loc_list temp_memory;
	s_mem_loc_list temp_heap;
	int i, j, k;

	//printf ("generate hash table entry\n");
	#define MAX_NUM_VALUES (MAX_STATE_SIZE * MAX_NUM_PROCESS) 
	for (i=0; i<MAX_NUM_VALUES; i++) {
		hash_table_entry[i] = -1;
	}
	
	i=0;
	for (j=0; j<MAX_NUM_PROCESS; j++) {
		k=0;
		temp_current_process = process_array[j];
		if (temp_current_process != NULL) {
			temp_curr_state = temp_current_process->process_state;
			if (temp_curr_state) {
				while (k<MAX_RETURN_STACK_SIZE) {
					hash_table_entry[i] = temp_curr_state->return_stack[k];
					i++; k++;
				}
				if (temp_current_process->curr_proc != NULL)
					hash_table_entry[i] = temp_current_process->curr_proc->proc_number;
				else 
					hash_table_entry[i] = -1;
				i++; k++;
				if (temp_current_process->curr_cfg_node != NULL)
					hash_table_entry[i] = temp_current_process->curr_cfg_node->node_number;
				else 
					hash_table_entry[i] = -1;
				i++; k++;
				temp_heap = temp_curr_state->heap;	
				while (temp_heap != NULL) {
					if (temp_heap->location->val != NULL) {
						if (temp_heap->location->val->kind == INT_KIND) {
							hash_table_entry[i] = temp_heap->location->val->value.int_val;
							i++; k++;
						}
					}
					temp_heap = temp_heap->next;
				}
				temp_memory = temp_curr_state->memory;	
				while (temp_memory != NULL) {
					if (temp_memory->location->val != NULL) {
						if (temp_memory->location->val->kind == INT_KIND) {
							hash_table_entry[i] = temp_memory->location->val->value.int_val;
							i++; k++;
						}
					}
					temp_memory = temp_memory->next;
				}
			}
			temp_heap = NULL;
			temp_memory = NULL;
		}
		if (k<MAX_STATE_SIZE) {
			while (k<MAX_STATE_SIZE) {
				i++; k++;
			}
		}
	}
	length_of_lasso++;

	/*For dining philosophers*/
	#ifdef RUN_PHIL
	{
		bool bool_property = 0;
		for (j=0; j<MAX_NUM_PROCESS; j++) {
			temp_current_process = process_array[j];	
			bool_property = 1;
			if (temp_current_process != NULL) {
				if ((bool_property == 1) && (temp_current_process->curr_proc && temp_current_process->curr_proc->proc_number == 2)) {
					bool_property = 1;
				} else 
					return; 
				if ((bool_property == 1)&& (temp_current_process->curr_cfg_node && temp_current_process->curr_cfg_node->node_number == 2)) {
					bool_property = 1;
				} else 
					return; 
			}
		}
		temp_heap = r_curr_state->heap;	
		while (temp_heap != NULL) {
			if (temp_heap->location->val != NULL) {
				if (temp_heap->location->val->kind == INT_KIND) {
					if ((bool_property == 1) && (temp_heap->location->val->value.int_val == 0)){
						bool_property = 1;
					} else 
						return; 
				}
			}
			temp_heap = temp_heap->next;
		}
		if (bool_property == 1) {
			printf ("All the processes are in deadlock \n");
			prev_N_counter_examples = N_counter_examples;
			N_counter_examples++;
			exit(0); //For Geometric, as we stop after 1st counter example 
			return; //For Quantitative, as we do not stop after 1st counter example 
		}
	}
	#endif
	//printf ("finished generate hash table entry\n");
	return;
}


void free_state (s_state * state) {
	s_mem_loc_list temp_memory;
	s_mem_loc_list prev_memory;
	s_mem_loc_list temp_heap;
	s_mem_loc_list prev_heap;
	
	//if (!state || state->freed == 1) {
	if (state->freed == 1) {
		return;
	}
	temp_memory = state->memory;
	while (temp_memory != NULL) {
		if (temp_memory->location->val) {
			free (temp_memory->location->val);
			mf_memory = mf_memory + sizeof (s_value);
			temp_memory->location->val = NULL;
		}
		if (temp_memory->location) {
			free (temp_memory->location);
			mf_memory = mf_memory + sizeof (s_mem_loc);
			temp_memory->location = NULL;
		}
		prev_memory = temp_memory;
		temp_memory = temp_memory->next;
		free (prev_memory);
		prev_memory = NULL;
		mf_memory = mf_memory + sizeof (struct my_mem_list);
	}
	temp_heap = state->heap;
	while (temp_heap != NULL) {
		if (temp_heap->location->val) {
			free (temp_heap->location->val);
			mf_memory = mf_memory + sizeof (s_value);
			temp_heap->location->val =NULL;
		}
		if (temp_heap->location) {
			free (temp_heap->location);
			mf_memory = mf_memory + sizeof (s_mem_loc);
			temp_heap->location = NULL;
		}
		prev_heap = temp_heap;
		temp_heap = temp_heap->next;
		free (prev_heap);
		mf_memory = mf_memory + sizeof (struct my_mem_list);
		prev_heap = NULL;
	}
	//printf ("h7 %d\n", temp_heap);
	state->freed = 1;
	//free (state);
	//mf_memory = mf_memory + sizeof (s_state);
	state = NULL;
	//printf ("freed state \n");
	return;
}


void push_state_in_stack (s_state * state) {
	struct state_list * temp_state_list;
	temp_state_list = (struct state_list *) xmalloc (sizeof (struct state_list));
	mm_memory = mm_memory + sizeof (struct state_list);
	temp_state_list->state = state;
	//printf ("pushed state \n");
	temp_state_list->next = NULL;
	if (state_stack == NULL) {
		state_stack = temp_state_list;	
	} else {
		temp_state_list->next = state_stack;
		state_stack = temp_state_list;
	}
	return;
}

struct state_list * pop_state_from_stack () {
	struct state_list * temp_state_list;
	if (state_stack == NULL) {
		return NULL;
	} else {
		temp_state_list = state_stack;
		state_stack = state_stack->next;
		//printf ("popped state %d\n", temp_state_list->state->number);
		return temp_state_list;
	}
}
