
/*========================================================================
               Copyright (C) 2004 by Xiaowan Huang
                            All rights reserved

    Permission is hereby granted, without written agreement and without
    license or royalty fees, to use, reproduce, prepare derivative
    works, distribute, and display this software and its documentation
    for any purpose, provided that (1) the above copyright notice and
    the following two paragraphs appear in all copies of the source code
    and (2) redistributions, including without limitation binaries,
    reproduce these notices in the supporting documentation. Substantial
    modifications to this software may be copyrighted by their authors
    and need not follow the licensing terms described here, provided
    that the new terms are clearly indicated in all files where they apply.
=======================================================================*/

/*************************************************************************
  FILE:  concur.c 
  DESCR: 
  AUTH:  Xiaowan Huang
  DATE:  (C) Thu Dec 16 19:49:33 2004
*************************************************************************/

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/msg.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <alloca.h>
#include <errno.h>
#include <assert.h>

#include "concur.h"

#define CONCUR_SEM_KEY	(key_t)10221

concur_ctrl_t *concur_ctrl_init(const char *socket_path)
{
	concur_ctrl_t *ctrl = ALLOC(concur_ctrl_t, 1);
	bzero(&ctrl->cli_addr, sizeof(struct sockaddr_un));
	ctrl->cli_addr.sun_family = AF_UNIX;
	sprintf(ctrl->cli_addr.sun_path, "%s%d", socket_path, getpid());
	printf("address: %s\n", ctrl->cli_addr.sun_path);
	
	ctrl->ctrl_sock = socket(AF_UNIX, SOCK_DGRAM, 0);
	DIE_ON_ERROR(ctrl->ctrl_sock, "socket");
	unlink(ctrl->cli_addr.sun_path);
	DIE_ON_ERROR(bind(ctrl->ctrl_sock,(struct sockaddr *)&ctrl->cli_addr, 
		sizeof(struct sockaddr_un)), "bind");
	
	bzero(&ctrl->serv_addr, sizeof(struct sockaddr_un));
	ctrl->serv_addr.sun_family = AF_UNIX;
	strcpy(ctrl->serv_addr.sun_path, socket_path);

	ctrl->semid = semget(CONCUR_SEM_KEY, 1, 0777);
	DIE_ON_ERROR(ctrl->semid, "semget");
	return ctrl;
}

void concur_ctrl_finalize(concur_ctrl_t *ctrl)
{
	if (ctrl->ctrl_sock > 0)
		close(ctrl->ctrl_sock);
	FREE(ctrl);
}

int concur_register_proc(const concur_ctrl_t *ctrl)
{
	ctrl_msg_t ctrl_msg;
	ctrl_msg.type = REGISTER_PROC;
	ctrl_msg.ctrl = REQUEST;
	ctrl_msg.pid = getpid();
	ctrl_msg.val = 0;
	ctrl_msg.error = 0;

	ssize_t count;
	count = sendto(ctrl->ctrl_sock, &ctrl_msg, sizeof(ctrl_msg_t), 0, 
			(struct sockaddr *)&ctrl->serv_addr, 
			sizeof(struct sockaddr_un));
	DIE_ON_ERROR(count, "sendto");
	printf("sent %d bytes\n", count);
	
	count = recvfrom(ctrl->ctrl_sock, &ctrl_msg, sizeof(ctrl_msg_t),
			0, NULL, 0);
	DIE_ON_ERROR(count, "recvfrom");
	printf("received %d bytes\n", count);

	if (ctrl_msg.ctrl != GRANTED) { 
		errno = ctrl_msg.error;
		return -1;
	}
	return 0;
}

int concur_unregister_proc(const concur_ctrl_t *ctrl)
{
	ctrl_msg_t ctrl_msg;
	ctrl_msg.type = UNREGISTER_PROC;
	ctrl_msg.ctrl = REQUEST;
	ctrl_msg.pid = getpid();
	ctrl_msg.val = 0;
	ctrl_msg.error = 0;

	ssize_t count;
	count = sendto(ctrl->ctrl_sock, &ctrl_msg, sizeof(ctrl_msg_t), 0, 
			(struct sockaddr *)&ctrl->serv_addr, 
			sizeof(struct sockaddr_un));
	DIE_ON_ERROR(count, "sendto");
	
	count = recvfrom(ctrl->ctrl_sock, &ctrl_msg, sizeof(ctrl_msg_t),
			0, NULL, 0);
	DIE_ON_ERROR(count, "recvfrom");

	if (ctrl_msg.ctrl != GRANTED) { 
		errno = ctrl_msg.error;
		return -1;
	}
	return 0;
}

static int make_word(int hi, int low)
{
	return (hi << 16) | low; 
}

// return the id of new message queue
// <0 on error
int concur_create_mqueue(const concur_ctrl_t *ctrl, int mlen, int msize, int flag)
{
	ctrl_msg_t ctrl_msg;
	ctrl_msg.type = CREATE_MSGQUEUE;
	ctrl_msg.ctrl = REQUEST;
	ctrl_msg.pid = getpid();
	ctrl_msg.val = make_word(mlen, msize);
	ctrl_msg.flag = flag;

	ssize_t count;
	count = sendto(ctrl->ctrl_sock, &ctrl_msg, sizeof(ctrl_msg_t), 0,
			(struct sockaddr *)&ctrl->serv_addr, 
			sizeof(struct sockaddr_un));
	DIE_ON_ERROR(count, "sendto");

	count = recvfrom(ctrl->ctrl_sock, &ctrl_msg, sizeof(ctrl_msg_t), 0, NULL, 0);
	DIE_ON_ERROR(count, "recvfrom");

	if (ctrl_msg.ctrl != GRANTED) {
		errno = ctrl_msg.error;
		return -1;
	}
	return (int)ctrl_msg.val;
}

int concur_destroy_mqueue(const concur_ctrl_t *ctrl, int mqid)
{
	ctrl_msg_t ctrl_msg;
	ctrl_msg.type = DESTROY_MSGQUEUE;
	ctrl_msg.ctrl = REQUEST;
	ctrl_msg.pid = getpid();
	ctrl_msg.val = mqid;
	ctrl_msg.flag = 0;

	ssize_t count;
	count = sendto(ctrl->ctrl_sock, &ctrl_msg, sizeof(ctrl_msg_t), 0,
			(struct sockaddr *)&ctrl->serv_addr, 
			sizeof(struct sockaddr_un));
	DIE_ON_ERROR(count, "sendto");

	count = recvfrom(ctrl->ctrl_sock, &ctrl_msg, sizeof(ctrl_msg_t), 0, NULL, 0);
	DIE_ON_ERROR(count, "recvfrom");

	if (ctrl_msg.ctrl != GRANTED) {
		errno = ctrl_msg.error;
		return -1;
	}
	return 0;
}

static int P_operation(int semid)
{
	struct sembuf sbuf[1];
	sbuf[0].sem_num = 0;
	sbuf[0].sem_op = -1;
	sbuf[0].sem_flg = 0;
	return semop(semid, sbuf, 1);
}

static int V_operation(int semid)
{
	struct sembuf sbuf[1];
	sbuf[0].sem_num = 0;
	sbuf[0].sem_op = 1;
	sbuf[0].sem_flg = 0;
	return semop(semid, sbuf, 1);
}

struct _message {
    long mtype; 	//The message type.
    char mtext [1];	//The message is of length MSGSZ.
};

int concur_send_msg(const concur_ctrl_t *ctrl, int mqid, const void *ptr, size_t nbytes, int flag)
{
	DIE_ON_ERROR(P_operation(ctrl->semid), "P_op");
	
	struct _message *mbuf;
	mbuf = (struct _message *)alloca(nbytes + sizeof(int));
	mbuf->mtype = mqid; 	// use msqid as msg type
	memcpy(mbuf->mtext, ptr, nbytes);
	ssize_t count = msgsnd(mqid, mbuf, nbytes, IPC_NOWAIT);
	int errno_bak = errno;
	
	if (count < 0 && errno == EAGAIN) {
		concur_report_blocked(ctrl);
		DIE_ON_ERROR(V_operation(ctrl->semid), "V_op");
		// bug here, process can be interrupted.
		count = msgsnd(mqid, mbuf, nbytes, 0);
		concur_report_unblocked(ctrl);
		errno_bak = errno;
	}
	else { 
		DIE_ON_ERROR(V_operation(ctrl->semid), "V_op");
	}
	errno = errno_bak;
	return count;
}

int concur_recv_msg(const concur_ctrl_t *ctrl, int mqid, void *ptr, size_t nbytes, int flag)
{
	DIE_ON_ERROR(P_operation(ctrl->semid), "P_op");

	struct _message *mbuf;
	mbuf = (struct _message *)alloca(nbytes + sizeof(int));
	mbuf->mtype = mqid;
	ssize_t count = msgrcv(mqid, mbuf, nbytes, mqid, IPC_NOWAIT);
	int errno_bak = errno;
	
	if (count < 0 && errno == ENOMSG) {
		concur_report_blocked(ctrl);
		DIE_ON_ERROR(V_operation(ctrl->semid), "V_op");
		count = msgrcv(mqid, mbuf, nbytes, mqid, 0);
		// bug here, process can be interrupted.
		concur_report_unblocked(ctrl);
		errno_bak = errno;
	}
	else {
		DIE_ON_ERROR(V_operation(ctrl->semid), "V_op");
	}
	
	memcpy(ptr, mbuf->mtext, nbytes);
	errno = errno_bak;
	return count;
}


int concur_report_blocked(const concur_ctrl_t *ctrl)
{
	ctrl_msg_t ctrl_msg;
	ctrl_msg.type = REPORT_BLOCKED;
	ctrl_msg.ctrl = REQUEST;
	ctrl_msg.pid = getpid();
	ctrl_msg.val = 0;
	ctrl_msg.flag = 0;

	ssize_t count;
	count = sendto(ctrl->ctrl_sock, &ctrl_msg, sizeof(ctrl_msg_t), 0, 
			(struct sockaddr *)&ctrl->serv_addr, 
			sizeof(struct sockaddr_un));
	DIE_ON_ERROR(count, "sendto");
	return count;
}

int concur_report_unblocked(const concur_ctrl_t *ctrl)
{
	ctrl_msg_t ctrl_msg;
	ctrl_msg.type = REPORT_UNBLOCKED;
	ctrl_msg.ctrl = REQUEST;
	ctrl_msg.pid = getpid();
	ctrl_msg.val = 0;
	ctrl_msg.flag = 0;

	ssize_t count;
	count = sendto(ctrl->ctrl_sock, &ctrl_msg, sizeof(ctrl_msg_t), 0, 
			(struct sockaddr *)&ctrl->serv_addr, 
			sizeof(struct sockaddr_un));
	DIE_ON_ERROR(count, "sendto");
	return count;
}

