
/*========================================================================
               Copyright (C) 2004 by Xiaowan Huang 
                            All rights reserved

    Permission is hereby granted, without written agreement and without
    license or royalty fees, to use, reproduce, prepare derivative
    works, distribute, and display this software and its documentation
    for any purpose, provided that (1) the above copyright notice and
    the following two paragraphs appear in all copies of the source code
    and (2) redistributions, including without limitation binaries,
    reproduce these notices in the supporting documentation. Substantial
    modifications to this software may be copyrighted by their authors
    and need not follow the licensing terms described here, provided
    that the new terms are clearly indicated in all files where they apply.
=======================================================================*/

/*************************************************************************
  FILE:  concur.h
  DESCR: 
  AUTH:  Xiaowan Huang
  EMAIL: xhuang@cs.sunysb.edu
  DATE:  (C) Thu Dec 16 19:49:32 2004
*************************************************************************/


#ifndef _CONCUR_H_
#define _CONCUR_H_


#ifdef __cplusplus
extern "C" {
#endif

#include <sys/types.h>
#include <sys/un.h>

#define DIE_ON_ERROR(retval, func_name) \
	if (retval < 0) { \
		perror(func_name); \
		exit(-1); \
	}

#define ALLOC(type, len) \
	(type *)malloc(sizeof(type) * len)

#define FREE(ptr) \
	free(ptr)

#define CONCUR_SUCCEEDED 0
#define CONCUR_FAILED -1

#define CONCUR_MSGQ_KEY	(key_t)10220
#define CONCUR_SEM_KEY	(key_t)10221

typedef enum ctrl_msg_type_t {
	NONE = 0,
	REGISTER_PROC	= 1,
	UNREGISTER_PROC = 2,
	CREATE_MSGQUEUE = 3,
	DESTROY_MSGQUEUE = 4,
	REPORT_BLOCKED	= 5,
	REPORT_UNBLOCKED = 6,

	REQUEST		= 1000,
	GRANTED 	= 1001,
	REJECTED 	= 1002,
} ctrl_msg_type_t;

typedef struct ctrl_msg_t {
	ctrl_msg_type_t type;
	ctrl_msg_type_t ctrl;
	pid_t 	pid;
	int	val;
	union {
		int error;
		int flag;
	};
} ctrl_msg_t;

typedef struct concur_ctrl_t {
	struct sockaddr_un cli_addr;
	struct sockaddr_un serv_addr;
	int ctrl_sock;
	int semid;
} concur_ctrl_t;

/*
concur_manager_t *concur_manager_init();
void concur_manager_finalize(concur_manager_t *mgr);
*/

concur_ctrl_t *concur_ctrl_init(const char *socket_path);
void concur_ctrl_finalize(concur_ctrl_t *ctrl);

int concur_register_proc(const concur_ctrl_t *ctrl);
int concur_unregister_proc(const concur_ctrl_t *ctrl);

int concur_create_mqueue(const concur_ctrl_t *ctrl, int mlen, int msize, int flag);
int concur_destroy_mqueue(const concur_ctrl_t *ctrl, int mqid);

int concur_send_msg(const concur_ctrl_t *ctrl, int mqid, const void *ptr, size_t nbytes, int flag);
int concur_recv_msg(const concur_ctrl_t *ctrl, int mqid, void *ptr, size_t nbytes, int flag);
int concur_report_blocked(const concur_ctrl_t *ctrl);
int concur_report_unblocked(const concur_ctrl_t *ctrl);

#ifdef __cplusplus
}
#endif

#endif /* _CONCUR_H_ */

