
/*========================================================================
               Copyright (C) 2004 by Xiaowan Huang
                            All rights reserved

    Permission is hereby granted, without written agreement and without
    license or royalty fees, to use, reproduce, prepare derivative
    works, distribute, and display this software and its documentation
    for any purpose, provided that (1) the above copyright notice and
    the following two paragraphs appear in all copies of the source code
    and (2) redistributions, including without limitation binaries,
    reproduce these notices in the supporting documentation. Substantial
    modifications to this software may be copyrighted by their authors
    and need not follow the licensing terms described here, provided
    that the new terms are clearly indicated in all files where they apply.
=======================================================================*/

/*************************************************************************
  FILE:  conlib.c 
  DESCR: 
  AUTH:  Xiaowan Huang
  DATE:  (C) Mon Dec 20 15:47:28 2004
*************************************************************************/

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/msg.h>
#include <errno.h>

#include "concur.h"
#include "conlib.h"

concur_manager_t *concur_manager_init(const char *socket_path)
{
	int i;
	concur_manager_t *mgr = ALLOC(concur_manager_t, 1);

	mgr->size_procs = INIT_SIZE_PROCS;
	mgr->n_procs = 0;
	mgr->n_blocked = 0;
	mgr->size_mqueues = INIT_SIZE_MQUEUES;
	mgr->n_mqueues = 0;
	mgr->to_quit = 0;
	
	mgr->procs = ALLOC(proc_t *, INIT_SIZE_PROCS);
	for (i=0; i<INIT_SIZE_PROCS; ++i) {
		mgr->procs[i] = NULL;
	}
	
	mgr->mqueues = ALLOC(mqueue_t *, INIT_SIZE_MQUEUES);
	for (i=0; i<INIT_SIZE_MQUEUES; ++i) {
		mgr->mqueues[i] = NULL;
	}
/*
	mgr->semid_proc = semget(IPC_PRIVATE, 1, IPC_CREAT);
	DIE_ON_ERROR(mgr->semid_proc, "semget");

	mgr->semid_mqueue = semget(IPC_PRIVATE, 1, IPC_CREAT);
	DIE_ON_ERROR(mgr->semid_mqueue, "semget");
*/
	mgr->lstn_sock = socket(AF_UNIX, SOCK_DGRAM, 0);
	DIE_ON_ERROR(mgr->lstn_sock, "socket");
	bzero(&mgr->serv_addr, sizeof(struct sockaddr_un));
	mgr->serv_addr.sun_family = AF_UNIX;
	strcpy(mgr->serv_addr.sun_path, socket_path);
	unlink(socket_path);
	bind(mgr->lstn_sock, (struct sockaddr *)&mgr->serv_addr, 
		sizeof(struct sockaddr_un));

	mgr->semid = semget(CONCUR_SEM_KEY, 1, IPC_CREAT | 0777);
	DIE_ON_ERROR(mgr->semid, "semget");
	semctl(mgr->semid, 0, SETVAL, 1);
	return mgr;
}

void concur_manager_cleanup(concur_manager_t *mgr)
{
	int i;
	for (i=0; i<mgr->size_procs; ++i) {
		if (mgr->procs[i] != NULL) {
			FREE(mgr->procs[i]);
			mgr->procs[i] = NULL;
		}
	}

	for (i=0; i<mgr->size_mqueues; ++i) {
		if (mgr->mqueues[i] != NULL) {
			FREE(mgr->mqueues[i]);
			mgr->mqueues[i] = NULL;
		}
	}

	/*
	DIE_ON_ERROR(semctl(mgr->semid_mqueue, 0, IPC_RMID), "semctl");
	mgr->semid_mqueue = -1;
	DIE_ON_ERROR(semctl(mgr->semid_proc, 0, IPC_RMID), "semctl");
	mgr->semid_proc = -1;
	*/
	close(mgr->lstn_sock);
}

void concur_manager_finalize(concur_manager_t *mgr)
{
	concur_manager_cleanup(mgr);
	FREE(mgr);
}


void concur_print_manager_info(const concur_manager_t *mgr)
{
	static const char *status_str[] = {
		"running",
		"blocked",
		"dead"
	};
	int i;
	for (i=0; i<mgr->size_procs; ++i) {
		const proc_t *proc = mgr->procs[i];
		if (proc != NULL) {
			printf("Process %d: %s\n", proc->pid, status_str[proc->status]);
		}
	}
	printf("%d message queues remaining\n", mgr->n_mqueues);
}

/****************************************************************************/

int concur_add_proc(concur_manager_t *mgr, pid_t pid)
{
	int i, found;
	
	for (i=0, found=0; i<mgr->n_procs; ++i) {
		if (mgr->procs[i] != NULL && mgr->procs[i]->pid == pid) {
			found = 1;
			break;
		}
	}
	
	if (found) {
		return -EEXIST;
	}

	for (i=0, found=0; i<mgr->size_procs; ++i) {
		if (mgr->procs[i] == NULL) {
			found = 1;
			break;
		}
	}
	if (found) {
		mgr->procs[i] = ALLOC(proc_t, 1);
		mgr->procs[i]->pid = pid;
		mgr->procs[i]->status = RUNNING;
		mgr->n_procs++;
	}
	else {
		assert(mgr->n_procs == mgr->size_procs);
		int size = mgr->size_procs;
		proc_t **procs = ALLOC(proc_t *, size + INIT_SIZE_PROCS);
		for (i=0; i<size; ++i) {
			procs[i] = mgr->procs[i];
		}
		for (i=size; i<size+INIT_SIZE_PROCS; ++i) {
			procs[i] = NULL;
		}
		procs[size] = ALLOC(proc_t, 1);
		procs[size]->pid = pid;
		procs[size]->status = RUNNING;
		FREE(mgr->procs);
		mgr->procs = procs;
		mgr->size_procs += INIT_SIZE_PROCS;
		mgr->n_procs++;
	}
	
	return 0;
}

int concur_remove_proc(concur_manager_t *mgr, pid_t pid)
{
	int i, found;
	
	for (i=0, found=0; i<mgr->n_procs; ++i) {
		if (mgr->procs[i] != NULL && mgr->procs[i]->pid == pid) {
			found = 1;
			break;
		}
	}
	
	if (!found) {
		return -ENOENT;
	}
	
	FREE(mgr->procs[i]);
	mgr->procs[i] = NULL;
	mgr->n_procs--;
	return 0;
}


// return mqueue id or <0 on error
int concur_add_mqueue(concur_manager_t *mgr, ssize_t mlen, int msize, int flag)
{
	/*
	if (mlen <= 0 || msize <= 0) {
		return -EINVAL;
	}
	
	int i, loc;
	for (i=0, loc=-1; i<mgr->size_mqueues; ++i) {
		if (mgr->mqueues[i] == NULL) {
			loc = i;
			break;
		}
	}
	
	if (loc < 0) {	
		// not found any empty slot
		assert(mgr->n_mqueues == mgr->size_mqueues);
		mqueue_t **mqueues = ALLOC(mqueue_t *, mgr->size_mqueues + INIT_SIZE_MQUEUES);
		int j;
		for (j=0; j<mgr->size_mqueues; ++j) {
			mqueues[j] = mgr->mqueues[j];
		}
		for (j=mgr->size_mqueues; j<mgr->size_mqueues+INIT_SIZE_MQUEUES; ++j) {
			mqueues[j] = NULL;
		}
		loc = mgr->size_mqueues;
		mgr->size_mqueues += INIT_SIZE_MQUEUES;
		FREE(mgr->mqueues);
		mgr->mqueues = mqueues;
	}

	assert(mgr->mqueues[loc] == NULL);
	
	mqueue_t *mq = ALLOC(mqueue_t, 1);
	mq->mlen = mlen;
	mq->msize = msize;
	mq->flag = 0;
	
	mgr->mqueues[loc] = mq;
	mgr->n_mqueues++;
	return loc;	
	*/
	if (mlen <= 0 || msize <= 0) {
		return -EINVAL;
	}

	int msqid = msgget(IPC_PRIVATE, IPC_CREAT | 0777);
	if (msqid < 0) 
		return -errno;
	struct msqid_ds ds;
	if (msgctl(msqid, IPC_STAT, &ds) < 0)
		return -errno;
	ds.msg_qbytes = mlen * msize;
	if (msgctl(msqid, IPC_SET, &ds) < 0)
		return -errno;
	return msqid;
}

int concur_remove_mqueue(concur_manager_t *mgr, int mqid)
{
	/*
	if (mqid < 0)
		return -EINVAL;
	if (mgr->mqueues[mqid] == NULL)
		return -ENOENT;
	
	FREE(mgr->mqueues[mqid]);
	mgr->mqueues[mqid] = NULL;
	mgr->n_mqueues--;
	return 0;
	*/

	if (msgctl(mqid, IPC_RMID, 0) < 0)
		return -errno;
	return 0;
}

static void concur_kill_all(const concur_manager_t *mgr)
{
	int i;
	for (i=0; i<mgr->size_procs; ++i) {
		if (mgr->procs[i] != NULL) {
			pid_t pid = mgr->procs[i]->pid;
			DIE_ON_ERROR(kill(pid, SIGINT), "kill");
		}
	}
}

int concur_block_proc(concur_manager_t *mgr, pid_t pid)
{
	int i, found;
	for (i=0, found=0; i<mgr->size_procs; ++i) {
		if (mgr->procs[i] != NULL && mgr->procs[i]->pid == pid) {
			found = 1;
			break;
		}
	}

	if (!found) {
		printf("No such process: %d\n", pid);
		errno = ENOENT;
		return -1;
	}
	assert(mgr->procs[i]->status == RUNNING);
	mgr->procs[i]->status = BLOCKED;
	mgr->n_blocked++;
	
	if (mgr->n_blocked == mgr->n_procs) {
		printf("ALL PROCESS BLOCKED!\n");
		concur_kill_all(mgr);
	}
	
	return 0;
}

int concur_unblock_proc(concur_manager_t *mgr, pid_t pid)
{
	int i, found;
	for (i=0, found=0; i<mgr->size_procs; ++i) {
		if (mgr->procs[i] != NULL && mgr->procs[i]->pid == pid) {
			found = 1;
			break;
		}
	}

	if (!found) {
		printf("No such process: %d\n", pid);
		errno = ENOENT;
		return -1;
	}

	assert(mgr->procs[i]->status == BLOCKED);
	mgr->procs[i]->status = RUNNING;
	mgr->n_blocked--;
	return 0;
}
