
/*========================================================================
               Copyright (C) 2004 by Xiaowan Huang 
                            All rights reserved

    Permission is hereby granted, without written agreement and without
    license or royalty fees, to use, reproduce, prepare derivative
    works, distribute, and display this software and its documentation
    for any purpose, provided that (1) the above copyright notice and
    the following two paragraphs appear in all copies of the source code
    and (2) redistributions, including without limitation binaries,
    reproduce these notices in the supporting documentation. Substantial
    modifications to this software may be copyrighted by their authors
    and need not follow the licensing terms described here, provided
    that the new terms are clearly indicated in all files where they apply.
=======================================================================*/

/*************************************************************************
  FILE:  conlib.h
  DESCR: 
  AUTH:  Xiaowan Huang
  EMAIL: xhuang@cs.sunysb.edu
  DATE:  (C) Mon Dec 20 15:44:21 2004
*************************************************************************/


#ifndef _CONLIB_H_
#define _CONLIB_H_


#ifdef __cplusplus
extern "C" {
#endif
	
#define INIT_SIZE_MQUEUES	8
#define INIT_SIZE_PROCS		4

typedef enum status_t {
	RUNNING = 0,
	BLOCKED = 1,
	DEAD	= 2,
} status_t;

typedef struct proc_t {
	pid_t 	pid;
	status_t status;
} proc_t;

#define MQPATH_MAXLEN	128

typedef struct mqueue_t {
//	int 	mqid;
	ssize_t	mlen;
	int		msize;
	int		flag;
} mqueue_t;
	
typedef struct concur_manager_t {
	int 	size_procs;
	int 	n_procs;
	int 	n_blocked;
	proc_t	**procs;
	//int	semid_proc;
	int 	size_mqueues;
	int 	n_mqueues;
	mqueue_t **mqueues;
	//int	semid_mqueue;
	struct sockaddr_un serv_addr;
	int 	lstn_sock;
	int	semid;
	int 	to_quit;
} concur_manager_t;

concur_manager_t *concur_manager_init(const char *socket_path);
void concur_manager_cleanup(concur_manager_t *mgr);
void concur_manager_finalize(concur_manager_t *mgr);
void concur_print_manager_info(const concur_manager_t *mgr);

#define CCMQ_READ	0x01
#define CCMQ_WRITE	0x02

int concur_add_proc(concur_manager_t *mgr, pid_t pid);
int concur_remove_proc(concur_manager_t *mgr, pid_t pid);
int concur_add_mqueue(concur_manager_t *mgr, ssize_t mlen, int msize, int flag);
int concur_remove_mqueue(concur_manager_t *mgr, int mqid);
int concur_block_proc(concur_manager_t *mgr, pid_t pid);
int concur_unblock_proc(concur_manager_t *mgr, pid_t pid);

#ifdef __cplusplus
}
#endif

#endif /* _CONLIB_H_ */

